import unittest
import json
from ibm_watson_machine_learning.preprocessing.data_join_pipeline import OBMPipelineGraphBuilder


baseline = ['// Data join graph', 'digraph {', 'node [color=lightblue2 style=filled]',
                '1 [label=" Parsed problem description" tooltip=""]', '2 [label=" Loading data" tooltip=""]', '1 -> 2',
                '3 [label=" Feature extraction" tooltip="Extracting features from the path:[main]. Checking 1 over 5 paths. Found 10 features."]',
                '2 -> 3', 'node [color=lightgray]', '30 [label=Identity]', '3 -> 30', 'node [color=lightgreen]',
                '300 [label=group_id]', '30 -> 300', 'node [color=lightblue2]', '300 -> Identity', 'Identity -> 8',
                'node [color=lightgreen]', '301 [label=group_customer_id]', '30 -> 301', 'node [color=lightblue2]',
                '301 -> Identity', 'Identity -> 8', 'node [color=lightgreen]', '302 [label=transaction_id]',
                '30 -> 302', 'node [color=lightblue2]', '302 -> Identity', 'Identity -> 8', 'node [color=lightgreen]',
                '303 [label=comments]', '30 -> 303', 'node [color=lightblue2]', '303 -> Identity', 'Identity -> 8',
                'node [color=lightgray]', '31 [label=TimeStamp]', '3 -> 31', 'node [color=lightgreen]',
                '310 [label=prefix_0_time]', '31 -> 310', 'node [color=lightblue2]', '310 -> day_of_week',
                'day_of_week -> 8', '310 -> minute', 'minute -> 8', '310 -> hour', 'hour -> 8', '310 -> day_of_month',
                'day_of_month -> 8', '310 -> day_of_year', 'day_of_year -> 8', '310 -> month', 'month -> 8',
                '4 [label=" Feature extraction" tooltip="Extracting feature from the path:[main](group_id)[purchases]. Checking 2 over 5 paths. Found 28 features."]',
                '2 -> 4', 'node [color=lightgray]', '40 [label=TimeStampSeriesCutoff]', '4 -> 40',
                'node [color=lightgreen]', '400 [label=time]', '40 -> 400', 'node [color=lightblue2]',
                '400 -> max_gap_to_cutoff', 'max_gap_to_cutoff -> 8', '400 -> count', 'count -> 8',
                '400 -> normalised_count', 'normalised_count -> 8', '400 -> recent_gap_to_cutoff',
                'recent_gap_to_cutoff -> 8', 'node [color=lightgray]', '41 [label=TimeSeriesCutoff]', '4 -> 41',
                'node [color=lightgreen]', '410 [label=price]', '41 -> 410', 'node [color=lightblue2]', '410 -> trend',
                'trend -> 8', '410 -> variance', 'variance -> 8', '410 -> recent', 'recent -> 8',
                '410 -> sliding_window', 'sliding_window -> 8', '410 -> mean', 'mean -> 8', '410 -> min', 'min -> 8',
                '410 -> max', 'max -> 8', '410 -> normalised_sum', 'normalised_sum -> 8', '410 -> sum', 'sum -> 8',
                '5 [label=" Feature extraction" tooltip="Extracting feature from the path:[main](transaction_id)[transactions]. Checking 3 over 5 paths. Found 13 features."]',
                '2 -> 5', 'node [color=lightgray]', '50 [label=NumberSet]', '5 -> 50', 'node [color=lightgreen]',
                '500 [label=product_id]', '50 -> 500', 'node [color=lightblue2]', '500 -> count', 'count -> 8',
                '500 -> variance', 'variance -> 8', '500 -> mean', 'mean -> 8', '500 -> min', 'min -> 8', '500 -> max',
                'max -> 8', '500 -> sum', 'sum -> 8', 'node [color=lightgray]', '51 [label=ItemSet]', '5 -> 51',
                'node [color=lightgreen]', '510 [label=description]', '51 -> 510', 'node [color=lightblue2]',
                '510 -> count', 'count -> 8', '510 -> distinct_count', 'distinct_count -> 8', 'node [color=lightgray]',
                '52 [label=ItemSetPattern]', '5 -> 52', 'node [color=lightgreen]', '520 [label=description]',
                '52 -> 520', 'node [color=lightblue2]', '520 -> item', 'item -> 8',
                '6 [label=" Feature extraction" tooltip="Extracting feature from the path:[main](transaction_id)[transactions](product_id)[products]. Checking 4 over 5 paths. Found 11 features."]',
                '2 -> 6', 'node [color=lightgray]', '60 [label=NumberSet]', '6 -> 60', 'node [color=lightgreen]',
                '600 [label=price]', '60 -> 600', 'node [color=lightblue2]', '600 -> count', 'count -> 8',
                '600 -> variance', 'variance -> 8', '600 -> mean', 'mean -> 8', '600 -> min', 'min -> 8', '600 -> max',
                'max -> 8', '600 -> sum', 'sum -> 8', 'node [color=lightgray]', '61 [label=ItemSet]', '6 -> 61',
                'node [color=lightgreen]', '610 [label=type]', '61 -> 610', 'node [color=lightblue2]', '610 -> count',
                'count -> 8', '610 -> distinct_count', 'distinct_count -> 8', 'node [color=lightgray]',
                '62 [label=ItemSetPattern]', '6 -> 62', 'node [color=lightgreen]', '620 [label=type]', '62 -> 620',
                'node [color=lightblue2]', '620 -> item', 'item -> 8',
                '7 [label=" Feature extraction" tooltip="Extracting feature from the path:[main](group_customer_id)[customers]. Checking 5 over 5 paths. Found 22 features."]',
                '2 -> 7', 'node [color=lightgray]', '70 [label=ItemSet]', '7 -> 70', 'node [color=lightgreen]',
                '700 [label=name]', '70 -> 700', 'node [color=lightblue2]', '700 -> count', 'count -> 8',
                '700 -> distinct_count', 'distinct_count -> 8', 'node [color=lightgreen]', '701 [label=address]',
                '70 -> 701', 'node [color=lightblue2]', '701 -> count', 'count -> 8', '701 -> distinct_count',
                'distinct_count -> 8', 'node [color=lightgray]', '71 [label=NumberSet]', '7 -> 71',
                'node [color=lightgreen]', '710 [label=age]', '71 -> 710', 'node [color=lightblue2]', '710 -> count',
                'count -> 8', '710 -> variance', 'variance -> 8', '710 -> mean', 'mean -> 8', '710 -> min', 'min -> 8',
                '710 -> max', 'max -> 8', '710 -> sum', 'sum -> 8', 'node [color=lightgreen]',
                '711 [label=number_children]', '71 -> 711', 'node [color=lightblue2]', '711 -> count', 'count -> 8',
                '711 -> variance', 'variance -> 8', '711 -> mean', 'mean -> 8', '711 -> min', 'min -> 8', '711 -> max',
                'max -> 8', '711 -> sum', 'sum -> 8', 'node [color=lightgreen]', '712 [label=income]', '71 -> 712',
                'node [color=lightblue2]', '712 -> count', 'count -> 8', '712 -> variance', 'variance -> 8',
                '712 -> mean', 'mean -> 8', '712 -> min', 'min -> 8', '712 -> max', 'max -> 8', '712 -> sum',
                'sum -> 8', 'node [color=lightgray]', '72 [label=ItemSetPattern]', '7 -> 72', 'node [color=lightgreen]',
                '720 [label=name]', '72 -> 720', 'node [color=lightblue2]', '720 -> item', 'item -> 8',
                'node [color=lightgreen]', '721 [label=address]', '72 -> 721', 'node [color=lightblue2]', '721 -> item',
                'item -> 8',
                '8 [label=" Feature selection" tooltip=" Found 66 features after feature selection with deduplicate consistent"]',
                '9 [label=" Output feature and model" tooltip="Writing 66 features"]', '8 -> 9', '}']


def compare_graph_to_baseline(g):
    list_str_g = g.__str__().split('\n')
    return [line.strip() for line in list_str_g] == baseline

class TestOBMGraph(unittest.TestCase):
    obm_json_path = './autoai/artifacts/group_customer/obm.json'
    pipeline_json = None
    builder = None

    @classmethod
    def setUp(cls) -> None:
        with open(cls.obm_json_path) as json_file:
            cls.obm_json = json.load(json_file)
            cls.pipeline_json = cls.obm_json['Pipeline']

        assert len(TestOBMGraph.pipeline_json) > 0

    def test_01_initialize_graph_object(self):
        TestOBMGraph.builder = OBMPipelineGraphBuilder(TestOBMGraph.obm_json)

        assert isinstance(TestOBMGraph.builder, OBMPipelineGraphBuilder)
        assert TestOBMGraph.pipeline_json == TestOBMGraph.builder.pipeline_json
        assert TestOBMGraph.builder.get_last_non_join_iteration() == 2
        assert TestOBMGraph.builder.get_selection_iteration() == 8
        assert TestOBMGraph.builder.get_join_iterations() == [3, 4, 5, 6, 7]
        print(TestOBMGraph.builder.graph)

    def test_02_build_graph(self):
        TestOBMGraph.builder.build_graph()
        assert compare_graph_to_baseline(TestOBMGraph.builder.graph) is True

    def test_03_visualize_graph(self):
        out = TestOBMGraph.builder.graph.render(view=True)
        assert out == 'Digraph.gv.pdf'
class TestOBMLalePipeline(unittest.TestCase):
    obm_json_path = './autoai/artifacts/agent_perf/obm.json'
    pipeline_json = None
    builder = None

    @classmethod
    def setUp(cls) -> None:
        with open(cls.obm_json_path) as json_file:
            cls.obm_json = json.load(json_file)

        assert len(TestOBMLalePipeline.obm_json) > 0

    def test_01_initialize_graph_object(self):
        TestOBMLalePipeline.builder = OBMPipelineGraphBuilder(TestOBMLalePipeline.obm_json)

        assert isinstance(TestOBMLalePipeline.builder, OBMPipelineGraphBuilder)

    def test_02_build_lale_pipeline(self):
        lale_pipeline_prefix = TestOBMLalePipeline.builder.build_preprocessing_prefix()
        assert lale_pipeline_prefix.name() == 'Relational'
        preprocessing_prefix = lale_pipeline_prefix._impl.operator
        assert preprocessing_prefix is not None
        assert len(preprocessing_prefix.steps()) > 1


if __name__ == '__main__':
    unittest.main()



import logging

import numpy as np
from sklearn.metrics import roc_auc_score, average_precision_score, log_loss
from sklearn.metrics import confusion_matrix, f1_score, precision_score, \
    recall_score, classification_report
from sklearn.metrics import explained_variance_score, mean_squared_error, \
    mean_squared_log_error, median_absolute_error, median_absolute_error, \
    mean_absolute_error, r2_score
from sklearn.utils.multiclass import unique_labels

logger = logging.getLogger(__name__)


def _f1_precision_recall(y_true, y_pred):
    """Compute the f1, precision, recall based on the average method. \
        Uses sklearn util to compute this metric.
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param average: Required for multiclass/multilabel targets.
    :type average: `str`
    :return: f1, precision, recall or array of float
    :rtype: `float`, `float`, `float`
    """
    metrics = {}
    round_digits = 2
    try:
        metrics['f1'] = round(
            f1_score(y_true, y_pred, zero_division=0), round_digits)
        metrics['precision'] = round(
            precision_score(y_true, y_pred, zero_division=0), round_digits)
        metrics['recall'] = round(recall_score(
            y_true, y_pred, zero_division=0), round_digits)

    except ValueError as ve:
        logger.exception(ve)
        logger.info(
            "Error occurred while collecting binary classification metircs.")
        labels = unique_labels(y_true, y_pred)
        if(len(labels) > 2):
            logger.info(len(labels) + " labels given, expected number 2")

    except Exception as ex:
        logger.exception(ex)

    return metrics


def get_binary_classification_metrics(y_true, y_pred, eval_metrics=[]):
    """Compute and package different metrics for binary classification, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param metrics: metrics requested by the user
    :type metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """
    metrics = {}
    round_digits = 2
    metrics = _f1_precision_recall(y_true, y_pred)
    try:
        labels = unique_labels(y_true)
        metrics['average precision'] = round(
            average_precision_score(y_true, y_pred), round_digits)
        if len(labels) > 1:
            metrics['roc auc'] = round(
                roc_auc_score(y_true, y_pred), round_digits)
            metrics['negative log loss'] = round(
                log_loss(y_true, y_pred), round_digits)
    except ValueError as ve:
        logger.exception(ve)
        logger.info(
            "Error occurred while collecting binary classification metircs.")
        labels = unique_labels(y_true, y_pred)
        if(len(labels) > 2):
            logger.error(len(labels) + " labels given, expected number 2")

    except Exception as ex:
        logger.exception(ex)

    # for metric in eval_metrics:
    # process each requested metric and send back the results.

    return metrics


def get_multi_label_classification_metrics(y_true, y_pred, eval_metrics=[]):
    """Compute and package different metrics for multi label classification, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param metrics: metrics requested by the user
    :type metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """
    metrics = {}
    round_digits = 2
    multilabel_average_options = ['micro', 'macro', 'weighted']

    for avg in multilabel_average_options:

        try:
            metrics['f1 ' +
                    avg] = round(f1_score(y_true, y_pred, average=avg, zero_division=0), round_digits)
            metrics['precision ' +
                    avg] = round(precision_score(y_true, y_pred, average=avg, zero_division=0), round_digits)
            metrics['recall ' +
                    avg] = round(recall_score(y_true, y_pred, average=avg, zero_division=0), round_digits)
        except Exception as ex:
            logger.exception(ex)

    return metrics


def get_eval_metrics_for_classificaton(y_true, y_pred, eval_metrics={}):
    """Compute and package different metrics for classification problem, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param eval_metrics: metrics requested by the user
    :type eval_metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """
    metrics = {}
    try:
        y_pred = (np.asarray(y_pred)).round()
        y_true = (np.asarray(y_true)).round()

        # reshape for tensorflow2.1 predictions
        if y_pred[0].shape != y_true[0].shape:  # FIXME
            logger.info("reshaping y_pred")
            y_pred = list(np.argmax(y_pred, axis=-1))

        # TODO: handle mix types labels {'multilabel-indicator', 'multiclass'}
        labels = unique_labels(y_true, y_pred)

        if len(labels) <= 2:
            metrics = get_binary_classification_metrics(
                y_true, y_pred, eval_metrics)
        else:
            metrics = get_multi_label_classification_metrics(
                y_true, y_pred, metrics)

    except Exception as ex:
        logger.exception(ex)
        logger.exception("Error occurred while evaluating metrics")

    return metrics


def get_eval_metrics_for_regression(y_true, y_pred, eval_metrics={}):
    """Compute and package different metrics for regression problem, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param eval_metrics: metrics requested by the user
    :type eval_metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """

    metrics = {}

    try:
        metrics['Negative root mean squared error'] = round(mean_squared_error(
            y_true, y_pred, squared=True), 2)
        metrics['Negative mean absolute error'] = round(mean_absolute_error(
            y_true, y_pred), 2)
        metrics['Negative root mean squared log error'] = round(mean_squared_log_error(
            y_true, y_pred), 2)
        metrics['Explained Variance'] = round(
            explained_variance_score(y_true, y_pred), 2)
        metrics['Negative mean squared error'] = round(mean_squared_error(
            y_true, y_pred, squared=False), 2)
        metrics['Negative mean squared log error'] = round(np.sqrt(mean_squared_log_error(
            y_true, y_pred)), 2)
        metrics['Negative median absolute error'] = round(median_absolute_error(
            y_true, y_pred), 2)
        metrics['R2'] = round(r2_score(y_true, y_pred), 2)
    except Exception as ex:
        logger.error("Exception occurred while calculating metrics.")
        logger.error(ex)

    return metrics



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import json

from .function_adapter import FunctionAdapter
from ibm_watson_machine_learning.libs.repo.swagger_client.rest import ApiException
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.function_artifact import FunctionArtifact
from ibm_watson_machine_learning.libs.repo.swagger_client.models.ml_assets_create_function_input import MlAssetsCreateFunctionInput
from ibm_watson_machine_learning.libs.repo.swagger_client.models import TagRepository
from ibm_watson_machine_learning.libs.repo.swagger_client.models import PatchOperationFunctions

logger = logging.getLogger('FunctionCollection')


class FunctionCollection:
    """
    Client operating on functions in repository service.

    :param str base_path: base url to Watson Machine Learning instance
    :param MLRepositoryApi repository_api: client connecting to repository rest api
    :param MLRepositoryClient client: high level client used for simplification and argument for constructors
    """
    def __init__(self, base_path, repository_api, client):

        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryClient
        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryApi

        if not isinstance(base_path, str) and not isinstance(base_path, unicode):
            raise ValueError('Invalid type for base_path: {}'.format(base_path.__class__.__name__))

        if not isinstance(repository_api, MLRepositoryApi):
            raise ValueError('Invalid type for repository_api: {}'.format(repository_api.__class__.__name__))

        if not isinstance(client, MLRepositoryClient):
            raise ValueError('Invalid type for client: {}'.format(client.__class__.__name__))

        self.base_path = base_path
        self.repository_api = repository_api
        self.client = client

    def all(self, expand=None, runtime_name=None, runtime_id=None):
        """
        Gets info about all functions which belong to this user.

        Not complete information is provided by all(). To get detailed information about function use get().

        :return: info about functions
        :rtype: list[FunctionArtifact]
        """
        all_functions = self.repository_api.repository_functions_list(self._get_expand_query_param_val(expand), runtime_name, runtime_id)
        list_functions_artifact = []
        if all_functions is not None:
            resr = all_functions.resources
            for iter1 in resr:
                list_functions_artifact.append(FunctionAdapter(iter1, self.client).artifact())
            return list_functions_artifact
        else:
            return []

    def get(self, artifact_id, expand = None):
        """
        Gets detailed information about function.

        :param str artifact_id: uid used to identify function
        :return: returned object has all attributes of FunctionArtifact
        :rtype: FunctionArtifact
        """
        logger.debug('Fetching information about function: {}'.format(artifact_id))

        if not isinstance(artifact_id, str) and not isinstance(artifact_id, unicode):
            raise ValueError('Invalid type for artifact_id: {}'.format(artifact_id.__class__.__name__))

        function_output = self.repository_api.repository_function_get(artifact_id, self._get_expand_query_param_val(expand))

        if function_output is not None:

            return FunctionAdapter(function_output, self.client).artifact()
        else:
            logger.debug('Function with guid={} not found'.format(artifact_id))
            raise ApiException('Function with guid={} not found'.format(artifact_id))

    def revisions(self, artifact_id, expand = None):
        """
        Gets all available revisions.

        :param str artifact_id: uid used to identify function
        :return: ???
        :rtype: list[str]
        """

        if not isinstance(artifact_id, str) and not isinstance(artifact_id, unicode):
            raise ValueError('Invalid type for artifact_id: {}'.format(artifact_id.__class__.__name__))

        logger.debug('Fetching information about function: {}'.format(artifact_id))

        function_output = self.repository_api.repository_list_function_revisions(artifact_id, self._get_expand_query_param_val(expand))

        list_function_revision_artifact = [FunctionArtifact]
        if function_output is not None:
            resr = function_output.resources
            for iter1 in resr:
                list_function_revision_artifact.append(FunctionAdapter(iter1, self.client).artifact())
            return list_function_revision_artifact
        else:
            logger.debug('Function with guid={} not found'.format(artifact_id))
            raise ApiException('Function with guid={} not found'.format(artifact_id))

    def revision(self, artifact_id, rev, expand):
        """
        Gets Function revision with given artifact_id and ver
        :param str artifact_id: uid used to identify function
        :param str rev: uid used to identify revision of function
        :return: FunctionArtifact(FunctionLoader) -- returned object has all attributes of FunctionArtifact
        """
        logger.debug('Fetching information about function revision: {}, {}'.format(artifact_id, rev))

        if not isinstance(artifact_id, str) and not isinstance(artifact_id, unicode):
            raise ValueError('Invalid type for artifact_id: {}'.format(artifact_id.__class__.__name__))

        if not isinstance(rev, str) and not isinstance(rev, unicode):
            raise ValueError('Invalid type for rev: {}'.format(rev.__class__.__name__))

        function_revision_output = self.repository_api.repository_get_function_revision(artifact_id, rev, self._get_expand_query_param_val(expand))
        if function_revision_output is not None:
            if function_revision_output is not None:
                return FunctionAdapter(function_revision_output[0], self.client).artifact()
            else:
                raise Exception('Function with guid={} not found'.format(artifact_id))
        else:
            raise Exception('Function with guid={} not found'.format(artifact_id))

    def revision_from_href(self, artifact_revision_href, expand = None):
        """
        Gets function revision from given href

        :param str artifact_revision_href: href identifying artifact and revision
        :return: returned object has all attributes of FunctionArtifact
        :rtype: FunctionArtifact(FunctionLoader)
        """

        if not isinstance(artifact_revision_href, str) and not isinstance(artifact_revision_href, unicode):
            raise ValueError('Invalid type for artifact_revision_href: {}'.format(artifact_revision_href.__class__.__name__))

        matched = re.search('.*/v4/functions/([A-Za-z0-9\-]+)/revisions/([A-Za-z0-9\-]+)',
                            artifact_revision_href)
        if matched is not None:
            artifact_id = matched.group(1)
            revision_id = matched.group(2)
            return self.revision(artifact_id, revision_id, expand)
        else:
            raise ValueError('Unexpected artifact revision href: {} format'.format(artifact_revision_href))

    def save(self, artifact):
        """
        Saves function in repository service.

        :param FunctionArtifact artifact: artifact to be saved in the repository service
        :return: saved artifact with changed MetaProps
        :rtype: FunctionArtifact
        """

        logger.debug('Creating a new function: {}'.format(artifact.name))

        if not issubclass(type(artifact), FunctionArtifact) :
            raise ValueError('Invalid type for artifact: {}'.format(artifact.__class__.__name__))

        if artifact.meta.prop(MetaNames.FUNCTIONS.REVISION_URL) is not None:
            raise ApiException(400, 'Invalid operation: save the same function artifact twice')

        try:
            function_id = artifact.uid
            if function_id is None:
                function_input = self._prepare_function_input(artifact)
                r = self.repository_api.repository_new_function(function_input)
                statuscode = r[1]
                if statuscode != 201:
                    logger.info('Error while creating function: no location header')
                    raise ApiException(404, 'No artifact location')

                function_artifact = self._extract_function_from_output(r)
                location = r[2].get('Location')
                location_match = re.search('.*/v4/functions/([A-Za-z0-9\\-]+)', location)

                if location_match is not None:
                    function_id = location_match.group(1)
                else:
                    logger.info('Error while creating function: no location header')
                    raise ApiException(404, 'No artifact location')
                artifact_with_guid = artifact._copy(function_id)
                revision_location = function_artifact.meta.prop(MetaNames.FUNCTIONS.REVISION_URL)
                revision_id = function_artifact.meta.prop(MetaNames.FUNCTIONS.REVISION)

                if revision_location is not None:
                    content_stream = artifact_with_guid.reader().read()
                    self.repository_api.upload_function_revision(function_id, revision_id, content_stream)
                    content_stream.close()
                    artifact_with_guid.reader().close()
                    return function_artifact
                else:
                    logger.info('Error while creating function revision: no location header')
                    raise ApiException(404, 'No artifact location')
            else:
                raise ApiException(404, 'Function not found')

        except Exception as e:
            logger.info('Error in function creation')
            import traceback
            print(traceback.format_exc())
            raise e

    def _extract_function_from_output(self, service_output):
        return FunctionAdapter(service_output[0], self.client).artifact()

    def patch(self, patch_input, latest_rev_url):
        matched = re.search('.*/v4/functions/([A-Za-z0-9\-]+)/revisions/([A-Za-z0-9\-]+)',
                            latest_rev_url)
        if matched is not None:
          function_id = matched.group(1)
          revision_id = matched.group(2)

        function_patch_input = self.prepare_function_patch_input(patch_input)
        function_patch_output = self.repository_api.repository_patch_functions_with_http_info(function_id, function_patch_input, revision_id)
        statuscode = function_patch_output[1]

        if statuscode != 200:
            logger.info('Error while patching function: no location header')
            raise ApiException(statuscode,"Error while patching function")

        if function_patch_output is not None:
            new_artifact = FunctionAdapter(function_patch_output[0], self.client).artifact()
        return new_artifact


    def remove(self, function_id):
        """
        Removes function with given id.

        :param str function_id: uid used to identify function
        """

        if not isinstance(function_id, str) and not isinstance(function_id, unicode):
            raise ValueError('Invalid type for function_id: {}'.format(function_id.__class__.__name__))

        if(function_id.__contains__("/v4/functions")):
            matched = re.search('.*/v4/functions/([A-Za-z0-9\-]+)', function_id)
            if matched is not None:
                function_id_value = matched.group(1)
                self.remove(function_id_value)
            else:
                raise ValueError('Unexpected function artifact href: {} format'.format(function_id))
        else:
            return self.repository_api.repository_delete_function(function_id)

    @staticmethod
    def prepare_function_patch_input(patch_input):
        patch_list = []
        if isinstance(patch_input, str):
            patch_input_list = json.loads(patch_input)
            if isinstance(patch_input_list, list):
                for iter1 in patch_input_list:
                    function_patch = PatchOperationFunctions(
                        op = iter1.get('op'),
                        path= iter1.get('path'),
                        value = iter1.get('value', None),
                        _from =iter1.get('from', None),
                    )
                    patch_list.append(function_patch)

                return patch_list

    @staticmethod
    def _prepare_function_input(artifact):
        function_input = MlAssetsCreateFunctionInput()
        function_input.name = artifact.name
        function_input.description = artifact.meta.prop(MetaNames.FUNCTIONS.DESCRIPTION)
        function_input.type = artifact.meta.prop(MetaNames.FUNCTIONS.TYPE)
        function_input.runtime_url = artifact.meta.prop(MetaNames.RUNTIMES.URL)

        input_data_schema = artifact.meta.prop(MetaNames.FUNCTIONS.INPUT_DATA_SCHEMA)
        output_data_schema = artifact.meta.prop(MetaNames.FUNCTIONS.OUTPUT_DATA_SCHEMA)
        sample_scoring_input = artifact.meta.prop(MetaNames.FUNCTIONS.SAMPLE_SCORING_INPUT)


        function_input.input_data_schema = input_data_schema
        function_input.output_data_schema = output_data_schema
        function_input.sample_scoring_input = sample_scoring_input
        if artifact.meta.prop(MetaNames.FUNCTIONS.TAGS) is not None:
            tags=artifact.meta.prop(MetaNames.FUNCTIONS.TAGS)
            tags_data_list = []
            if isinstance(tags, str):
                tags_list = json.loads(artifact.meta.prop(MetaNames.FUNCTIONS.TAGS))
                if isinstance(tags_list, list):
                    for iter1 in tags_list:
                        tags_data = TagRepository()
                        for key in iter1:
                            if key == 'value':
                                tags_data.value= iter1['value']
                            if key == 'description':
                                tags_data.description = iter1['description']
                        tags_data_list.append(tags_data)
                else:
                    raise ValueError("Invalid tag Input")
                function_input.tags =  tags_data_list

        return function_input

    @staticmethod
    def _get_expand_query_param_val(expand):
       if expand is True:
        expand_val = 'runtime'
       else:
        expand_val = None
       return expand_val


"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import logging

import numpy as np
from sklearn.metrics import roc_auc_score, average_precision_score, log_loss
from sklearn.metrics import confusion_matrix, f1_score, precision_score, \
    recall_score, classification_report
from sklearn.metrics import explained_variance_score, mean_squared_error, \
    mean_squared_log_error, median_absolute_error, median_absolute_error, \
    mean_absolute_error, r2_score
from sklearn.utils.multiclass import unique_labels

logger = logging.getLogger(__name__)


def _f1_precision_recall(y_true, y_pred):
    """Compute the f1, precision, recall based on the average method. \
        Uses sklearn util to compute this metric.
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param average: Required for multiclass/multilabel targets.
    :type average: `str`
    :return: f1, precision, recall or array of float
    :rtype: `float`, `float`, `float`
    """
    metrics = {}
    round_digits = 2
    try:
        metrics['f1'] = round(
            f1_score(y_true, y_pred, zero_division=0), round_digits)
        metrics['precision'] = round(
            precision_score(y_true, y_pred, zero_division=0), round_digits)
        metrics['recall'] = round(recall_score(
            y_true, y_pred, zero_division=0), round_digits)

    except ValueError as ve:
        logger.exception(ve)
        logger.info(
            "Error occurred while collecting binary classification metircs.")
        labels = unique_labels(y_true, y_pred)
        if(len(labels) > 2):
            logger.info(len(labels) + " labels given, expected number 2")

    except Exception as ex:
        logger.exception(ex)

    return metrics


def get_binary_classification_metrics(y_true, y_pred, eval_metrics=[]):
    """Compute and package different metrics for binary classification, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param metrics: metrics requested by the user
    :type metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """
    metrics = {}
    round_digits = 2
    metrics = _f1_precision_recall(y_true, y_pred)
    try:
        labels = unique_labels(y_true)
        metrics['average precision'] = round(
            average_precision_score(y_true, y_pred), round_digits)
        if len(labels) > 1:
            metrics['roc auc'] = round(
                roc_auc_score(y_true, y_pred), round_digits)
            metrics['negative log loss'] = round(
                log_loss(y_true, y_pred), round_digits)
    except ValueError as ve:
        logger.exception(ve)
        logger.info(
            "Error occurred while collecting binary classification metircs.")
        labels = unique_labels(y_true, y_pred)
        if(len(labels) > 2):
            logger.error(len(labels) + " labels given, expected number 2")

    except Exception as ex:
        logger.exception(ex)

    # for metric in eval_metrics:
    # process each requested metric and send back the results.

    return metrics


def get_multi_label_classification_metrics(y_true, y_pred, eval_metrics=[]):
    """Compute and package different metrics for multi label classification, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param metrics: metrics requested by the user
    :type metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """
    metrics = {}
    round_digits = 2
    multilabel_average_options = ['micro', 'macro', 'weighted']

    for avg in multilabel_average_options:

        try:
            metrics['f1 ' +
                    avg] = round(f1_score(y_true, y_pred, average=avg, zero_division=0), round_digits)
            metrics['precision ' +
                    avg] = round(precision_score(y_true, y_pred, average=avg, zero_division=0), round_digits)
            metrics['recall ' +
                    avg] = round(recall_score(y_true, y_pred, average=avg, zero_division=0), round_digits)
        except Exception as ex:
            logger.exception(ex)

    return metrics


def get_eval_metrics_for_classificaton(y_true, y_pred, eval_metrics={}):
    """Compute and package different metrics for classification problem, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param eval_metrics: metrics requested by the user
    :type eval_metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """
    metrics = {}
    try:
        y_pred = (np.asarray(y_pred)).round()
        y_true = (np.asarray(y_true)).round()

        # reshape for tensorflow2.1 predictions
        if y_pred[0].shape != y_true[0].shape:  # FIXME
            logger.info("reshaping y_pred")
            y_pred = list(np.argmax(y_pred, axis=-1))

        # TODO: handle mix types labels {'multilabel-indicator', 'multiclass'}
        labels = unique_labels(y_true, y_pred)

        if len(labels) <= 2:
            metrics = get_binary_classification_metrics(
                y_true, y_pred, eval_metrics)
        else:
            metrics = get_multi_label_classification_metrics(
                y_true, y_pred, metrics)

    except Exception as ex:
        logger.exception(ex)
        logger.exception("Error occurred while evaluating metrics")

    return metrics


def get_eval_metrics_for_regression(y_true, y_pred, eval_metrics={}):
    """Compute and package different metrics for regression problem, \
    and return a dictionary with metrics
    :param y_true: 1d array-like, with ground truth (correct target values)
    :type y_true: `array`
    :param y_pred: 1d array-like, estimation returned by classifier.
    :type y_pred: `array`
    :param eval_metrics: metrics requested by the user
    :type eval_metrics: list of metrics which needs to be sent back.
    :return: dictionary with metrics
    :rtype: `dict`
    """

    metrics = {}

    try:
        metrics['Negative root mean squared error'] = round(mean_squared_error(
            y_true, y_pred, squared=True), 2)
        metrics['Negative mean absolute error'] = round(mean_absolute_error(
            y_true, y_pred), 2)
        metrics['Negative root mean squared log error'] = round(mean_squared_log_error(
            y_true, y_pred), 2)
        metrics['Explained Variance'] = round(
            explained_variance_score(y_true, y_pred), 2)
        metrics['Negative mean squared error'] = round(mean_squared_error(
            y_true, y_pred, squared=False), 2)
        metrics['Negative mean squared log error'] = round(np.sqrt(mean_squared_log_error(
            y_true, y_pred)), 2)
        metrics['Negative median absolute error'] = round(median_absolute_error(
            y_true, y_pred), 2)
        metrics['R2'] = round(r2_score(y_true, y_pred), 2)
    except Exception as ex:
        logger.error("Exception occurred while calculating metrics.")
        logger.error(ex)

    return metrics



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
"""
Pickle based serialization
"""
import pickle
from ibmfl.message.message import Message
from ibmfl.message.serializer import Serializer


class PickleSerializer(Serializer):
    """
    Class for Pickle based serialization
    """

    def serialize(self, msg):
        """
        Serialize a message using pickle

        :param msg: message to serialize
        :type msg: `Message`
        :return: serialize byte stream
        :rtype: `b[]`
        """
        msg_header = msg.get_header()
        serialized_data = msg.get_data()  # need to serialize the data
        return pickle.dumps({'header': msg_header,
                             'data': serialized_data,
                             })

    def deserialize(self, serialized_byte_stream):
        """
        Deserialize a byte stream to a message

        :param serialized_byte_stream: byte stream
        :type serialized_byte_stream: `b[]`
        :return: deserialized message
        :rtype: `Message`
        """
        data_dict = pickle.loads(serialized_byte_stream)
        if 'MSG_LEN|' in data_dict:
            msg_length = int(data_dict.split('|')[1])
            return msg_length

        msg = Message(data=data_dict['data'])
        msg.set_header(data_dict['header'])

        return msg



#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
warnings.filterwarnings("ignore", category=FutureWarning)

from .ml_repository_artifact import MLRepositoryArtifact
from .spark_artifact_loader import SparkArtifactLoader
from .spark_pipeline_artifact import SparkPipelineArtifact
from .spark_pipeline_loader import SparkPipelineLoader
from .spark_pipeline_model_artifact import SparkPipelineModelArtifact
from .spark_pipeline_model_loader import SparkPipelineModelLoader
from .spark_pipeline_reader import SparkPipelineReader
from .spark_version import SparkVersion
from .version_helper import VersionHelper
from .libraries_artifact import LibrariesArtifact
from .libraries_artifact_loader import LibrariesArtifactLoader
from .libraries_artifact_reader import LibrariesArtifactReader
from .libraries_loader import LibrariesLoader
from .runtimes_artifact import RuntimesArtifact
from .runtimes_artifact_reader import RuntimesArtifactReader
from .runtimes_artifact_loader import RuntimesArtifactLoader
from .hybrid_pipeline_model_artifact import HybridPipelineModelArtifact
from .hybrid_artifact_loader import HybridArtifactLoader
from .hybrid_pipeline_model_loader import HybridPipelineModelLoader

from .content_loaders import SparkPipelineContentLoader, IBMSparkPipelineContentLoader, SparkPipelineModelContentLoader,\
    IBMSparkPipelineModelContentLoader, MLPipelineContentLoader, MLPipelineModelContentLoader
from .python_version import PythonVersion

__all__ = ['MLRepositoryArtifact', 'SparkArtifactLoader', 'SparkPipelineArtifact', 'SparkPipelineLoader',
           'SparkPipelineModelArtifact', 'SparkPipelineModelLoader', 'SparkPipelineReader', 'SparkVersion',
           'VersionHelper', 'SparkPipelineContentLoader', 'MLPipelineModelContentLoader',
           'IBMSparkPipelineContentLoader', 'SparkPipelineModelContentLoader', 'IBMSparkPipelineModelContentLoader',
           'MLPipelineContentLoader', 'PythonVersion', 'LibrariesArtifact', 'LibrariesArtifactLoader',
           'LibrariesArtifactReader', 'LibrariesLoader', 'RuntimesArtifact', 'RuntimesArtifactReader',
           'RuntimesArtifactLoader', 'HybridPipelineModelArtifact', 'HybridArtifactLoader', 'HybridPipelineModelLoader'
           ]



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""



#!/usr/bin/env python3
import os
import sys
import importlib.util

import requests
import json

def process_heartbeat(kwargs):
    try:
        config_dict = kwargs.get('config_dict')
        agg_info = config_dict.get('aggregator')
        wml_services_url = agg_info["ip"].split("/")[0]

        hearbeat_resp = requests.get("https://" + wml_services_url + "/wml_services/training/heartbeat", verify=False)
        hearbeat_resp_json = json.loads(hearbeat_resp.content.decode("utf-8"))
        print(hearbeat_resp_json.get("service_implementation"))
        service_version = hearbeat_resp_json.get("service_implementation")

        if service_version.startswith("4.0"):
            return "cpd402"
        if service_version.startswith("353"):
            return "cpd35"

        return "cloud"

    except Exception as ex:
        print("unable to process heartbeat")
        return "cloud"



def import_source(module_file_path):
    module_name = 'ibmfl'
    
    module_spec = importlib.util.spec_from_file_location(
        module_name, module_file_path)
    module = importlib.util.module_from_spec(module_spec)
    sys.modules[module_name] = module
    module_spec.loader.exec_module(module)


fl_path = os.path.abspath('.')
if fl_path not in sys.path:
    sys.path.append(fl_path)



class Party:
    """
    Party Wrapper for determining required version of FL Party to invoke.
    """

    SUPPORTED_PLATFORMS_MAP = {
        'cloud':"/cloud/ibmfl/__init__.py",
        'cpd35':"/cpd35/ibmfl/__init__.py",
        'cpd402':"/cpd402/ibmfl/__init__.py",
    } 

    def __init__(self, **kwargs):

        ibmfl_module = sys.modules['ibmfl'] 
        ibmfl_module_location_list = ibmfl_module.__path__
        
        # base location string, default to cloud location 
        ibmfl_module_location = ibmfl_module_location_list[0] 
        
        # get arg for platform
        calculated_platform_env = process_heartbeat(kwargs)
        platform_env = kwargs.get('env', calculated_platform_env)
        
        # process location 
        ibmfl_module_location = ibmfl_module_location + self.SUPPORTED_PLATFORMS_MAP.get(platform_env)
        
        del sys.modules['ibmfl']
        del sys.modules['ibmfl.party']
        del sys.modules['ibmfl.party.party']
        import_source(ibmfl_module_location )
        from ibmfl.party.party import Party
        self.Party = Party(**kwargs)
        self.connection = self.Party.connection

    def start(self):
        self.Party.start()

    def isStopped(self):
        return self.Party.connection.stopped


# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
import requests
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,DATA_ASSETS_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos
from ibm_watson_machine_learning.metanames import ScriptMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure, ForbiddenActionForGitBasedProject
import os

_DEFAULT_LIST_LENGTH = 50


class Script(WMLResource):
    """
    Store and manage your scripts assets.

    """
    ConfigurationMetaNames = ScriptMetaNames()
    """MetaNames for script Assets creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, script_uid):
        """
            Get script asset details.

            **Parameters**

            .. important::
                #. **script_uid**: Unique id  of script\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: Metadata of the stored script asset\n
                **return type**: dict\n

            **Example**

             >>> script_details = client.scripts.get_details(script_uid)

        """
        Script._validate_type(script_uid, u'script_uid', STR_TYPE, True)


        if not self._ICP:
            response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(script_uid), params=self._client._params(),
                                    headers=self._client._get_headers())
        else:
            response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(script_uid), params=self._client._params(),
                                      headers=self._client._get_headers(), verify=False)
        if response.status_code == 200:
            response = self._get_required_element_from_response(self._handle_response(200, u'get asset details', response))

            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                return response
            else:

                entity = response[u'entity']

                try:
                    del entity[u'script'][u'ml_version']
                except KeyError:
                    pass

                final_response = {
                    "metadata": response[u'metadata'],
                    "entity": entity
                }

                return final_response

        else:
            return self._handle_response(200, u'get asset details', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store(self, meta_props, file_path):
        """
                Creates a Scripts asset and uploads content to it.

                **Parameters**

                .. important::
                   #. **meta_props**:  Name to be given to the Scripts asset\n

                      **type**: str\n

                   #. **file_path**:  Path to the content file to be uploaded\n

                      **type**: str\n

                **Output**

                .. important::

                    **returns**: metadata of the stored Scripts asset\n
                    **return type**: dict\n

                **Example**

                 >>> metadata = {
                 >>>        client.script.ConfigurationMetaNamess.NAME: 'my first script',
                 >>>        client.script.ConfigurationMetaNames.DESCRIPTION: 'description of the script',
                 >>>        client.script.ConfigurationMetaNames.SOFTWARE_SPEC_UID: '0cdb0f1e-5376-4f4d-92dd-da3b69aa9bda'
                 >>>    }
                 >>>
                 >>> asset_details = client.scripts.store(meta_props=metadata,file_path="/path/to/file")

        """
        if self._client.project_type == 'local_git_storage':
            raise ForbiddenActionForGitBasedProject(reason="Storing Scripts is blocked for git based project.")


        WMLResource._chk_and_block_create_update_for_python36(self)
        #Script._validate_type(name, u'name', str, True)
        Script._validate_type(file_path, u'file_path', str, True)
        script_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        name, extension = os.path.splitext(file_path)

        response = self._create_asset(script_meta, file_path, extension)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:

            entity = response[u'entity']

            try:
                del entity[u'script'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

    def _create_asset(self, script_meta, file_path, extension='.py'):

        ##Step1: Create a data asset
        f = {'file': open(file_path, 'rb')}
        name = script_meta['metadata'][u'name']

        if script_meta['metadata'].get('description') is not None:
            desc = script_meta['metadata'][u'description']
        else:
            desc = ""

        if script_meta.get('software_spec_uid') is not None:
            if script_meta[u'software_spec_uid'] == "":
                raise WMLClientError("Failed while creating a script asset, SOFTWARE_SPEC_UID cannot be empty")
                return

        base_script_asset = {
            "description": "Script File",
            "fields": [
                {
                    "key": "language",
                    "type": "string",
                    "facet": False,
                    "is_array": False,
                    "search_path": "asset.type",
                    "is_searchable_across_types": True
                }
            ],
            "relationships": [
                {
                    "key": "software_spec.id",
                    "target_asset_type": "software_specification",
                    "on_delete_target": "IGNORE",
                    "on_delete": "IGNORE",
                    "on_clone_target": "IGNORE"
                }
            ],
            "name": "script",
            "version": 1
        }

        if extension == '.py':
            lang = 'python3'
        elif extension == '.R':
            lang = 'R'
        else:
            raise WMLClientError("This file type is not supported. It has to be either a python script(.py file ) or a "
                                 "R script")

        #check if the software spec specified is base or derived and
        # accordingly update the entity for asset creation
        # TODO WSD2.0
        #kaganesa: remove the below if and else. once v2/software_spec apis are available in WSD2.0,
        # everything inside if should work.
        if not self._client.WSD:
            sw_spec_details = self._client.software_specifications.get_details(script_meta[u'software_spec_uid'])

            if lang == 'R':
                rscript_sw_spec_id = self._client.software_specifications.get_id_by_name('default_r3.6')

                if rscript_sw_spec_id != script_meta[u'software_spec_uid']:
                    raise WMLClientError("For R scripts, only base software spec 'default_r3.6' is supported. Specify "
                                         "the id you get via "
                                         "self._client.software_specifications.get_id_by_name('default_r3.6')")

            if(sw_spec_details[u'entity'][u'software_specification'][u'type'] == 'base'):
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                    asset_meta = {
                        "metadata": {
                            "name": name,
                            "description": desc,
                            "asset_type": "script",
                            "origin_country": "us",
                            "asset_category": "USER"
                        },
                        "entity": {
                            "script": {
                                "language": {
                                    # "name": "python3"
                                    "name": lang
                    },
                                "software_spec": {
                                    "base_id": script_meta[u'software_spec_uid']
                                }
                             }
                            }
                    }
                else:
                     asset_meta = {
                         "metadata": {
                             "name": name,
                             "description": desc,
                             "asset_type": "script",
                             "origin_country": "us",
                             "asset_category": "USER"
                         },
                         "entity": {
                             "script": {
                                 "ml_version": "4.0.0",
                                 "language": {
                                     # "name": "python3"
                                     "name": lang
                     },
                                 "software_spec": {
                                     "base_id": script_meta[u'software_spec_uid']
                                 }
                             }
                         }
                     }
            else:
                if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                    asset_meta = {
                        "metadata": {
                            "name": name,
                            "description": desc,
                            "asset_type": "script",
                            "origin_country": "us",
                            "asset_category": "USER"
                        },
                        "entity": {
                            "script": {
                                "language": {
                                    # "name": "python3"
                                    "name": lang
                    },
                                "software_spec": {
                                    "id": script_meta[u'software_spec_uid']
                                }
                            }
                        }
                    }
                else:
                    asset_meta = {
                        "metadata": {
                            "name": name,
                            "description": desc,
                            "asset_type": "script",
                            "origin_country": "us",
                            "asset_category": "USER"
                        },
                        "entity": {
                            "script": {
                                "ml_version": "4.0.0",
                                "language": {
                                    # "name": "python3"
                                    "name": lang
                                },
                                "software_spec": {
                                    "id": script_meta[u'software_spec_uid']
                                }
                            }
                        }
                    }
        else:
            asset_meta = {
                "metadata": {
                    "name": name,
                    "description": desc,
                    "asset_type": "script",
                    "origin_country": "us",
                    "asset_category": "USER"
                },
                "entity": {
                    "script": {
                        "language": {
                            # "name": "python3"
                            "name": lang
            },
                        "software_spec": {
                            "id": script_meta[u'software_spec_uid']
                        }
                    }
                }
            }

        #Step1  : Create an asset
        print("Creating Script asset...")

        if self._client.WSD:
            # For WSD the asset creation is done within _wsd_create_asset function using polyglot
            # Thus using the same for data_assets type


            meta_props = {
                    "name": name
            }
            details = Script._wsd_create_asset(self, "script", asset_meta, meta_props, file_path, user_archive_file=True)
            return self._get_required_element_from_response(details)
        else:
            if not self._ICP:
                creation_response = requests.post(
                        self._client.service_instance._href_definitions.get_assets_href(),
                        headers=self._client._get_headers(),
                        params = self._client._params(),
                        json=asset_meta
                )
            else:
                if self._client.ICP_PLATFORM_SPACES:
                    creation_response = requests.post(
                        self._client.service_instance._href_definitions.get_data_assets_href(),
                        headers=self._client._get_headers(),
                        json=asset_meta,
                        params=self._client._params(),
                        verify=False
                    )
                else:
                    asset_type_response = requests.post(
                        self._client.service_instance._href_definitions.get_wsd_asset_type_href() + '?',
                        headers=self._client._get_headers(),
                        json=base_script_asset,
                        params=self._client._params(),
                        verify=False
                    )
                    if asset_type_response.status_code == 201 or asset_type_response.status_code == 409:
                        creation_response = requests.post(
                            self._client.service_instance._href_definitions.get_data_assets_href(),
                            headers=self._client._get_headers(),
                            json=asset_meta,
                            params=self._client._params(),
                            verify=False
                        )

            asset_details = self._handle_response(201, u'creating new asset', creation_response)
            #Step2: Create attachment
            if creation_response.status_code == 201:
                asset_id = asset_details["metadata"]["asset_id"]
                attachment_meta = {
                        "asset_type": "script",
                        "name": "attachment_"+asset_id
                    }

                if not self._ICP:
                    attachment_response = requests.post(
                        self._client.service_instance._href_definitions.get_attachments_href(asset_id),
                        headers=self._client._get_headers(),
                        params=self._client._params(),
                        json=attachment_meta
                    )
                else:
                    attachment_response = requests.post(
                        self._client.service_instance._href_definitions.get_attachments_href(asset_id),
                        headers=self._client._get_headers(),
                        json=attachment_meta,
                        params=self._client._params(),
                        verify=False
                    )
                attachment_details = self._handle_response(201, u'creating new attachment', attachment_response)
                if attachment_response.status_code == 201:
                    attachment_id = attachment_details["attachment_id"]
                    attachment_url = attachment_details["url1"]


                    #Step3: Put content to attachment
                    if not self._ICP:

                        files = open(file_path)

                        put_response = requests.put(
                            attachment_url,
                            data=files.read()
                        )
                    else:
                        put_response = requests.put(
                            self._wml_credentials['url'] + attachment_url,
                            files=f,
                            verify=False
                        )
                    if put_response.status_code == 201 or put_response.status_code == 200:
                        # Step4: Complete attachment
                        if not self._ICP:
                            complete_response = requests.post(
                                # self._client.service_instance._href_definitions.get_attachment_href(asset_id,attachment_id)+"/complete",
                                self._client.service_instance._href_definitions.get_attachment_complete_href(asset_id, attachment_id),
                                headers=self._client._get_headers(),
                                params = self._client._params()

                            )
                        else:
                            complete_response = requests.post(
                                self._client.service_instance._href_definitions.get_attachment_href(asset_id,attachment_id)+"/complete",
                                headers=self._client._get_headers(),
                                params=self._client._params(),
                                verify=False
                            )
                        if complete_response.status_code == 200:
                            print("SUCCESS")
                            return self._get_required_element_from_response(asset_details)
                        else:
                            self._delete(asset_id)
                            raise WMLClientError("Failed while creating a script asset. Try again.")
                    else:
                        self._delete(asset_id)
                        raise WMLClientError("Failed while creating a script asset. Try again.")
                else:
                    print("SUCCESS")
                    return self._get_required_element_from_response(asset_details)
            else:
                raise WMLClientError("Failed while creating a script asset. Try again.")


    def list(self, limit=None):
        """
           List stored scripts. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all script in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.script.list()
        """

        Script._validate_type(limit, u'limit', int, False)
        href = self._client.service_instance._href_definitions.get_search_script_href()

        data = {
                "query": "*:*"
        }
        if limit is not None:
            data.update({"limit": limit})

        if not self._ICP:
            response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(),json=data)
        else:
            response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(),json=data, verify=False)
        self._handle_response(200, u'list assets', response)
        asset_details = self._handle_response(200, u'list assets', response)["results"]
        space_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_type'], m["metadata"]["asset_id"]) for
            m in asset_details]

        self._list(space_values, [u'NAME', u'ASSET_TYPE', u'ASSET_ID'], None, _DEFAULT_LIST_LENGTH)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def download(self, asset_uid, filename, rev_uid=None):
        """
            Download the content of a script asset.

            **Parameters**

            .. important::

                 #. **asset_uid**:  The Unique Id of the script asset to be downloaded\n
                    **type**: str\n

                 #. **filename**:  filename to be used for the downloaded file\n
                    **type**: str\n

                 #. **rev_uid**:  Revision id\n
                    **type**: str\n

            **Output**

            .. important::

                 **returns**: Path to the downloaded asset content\n
                 **return type**: str\n

            **Example**

             >>> client.script.download(asset_uid,"script_file.zip")
         """
        if rev_uid is not None and self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not applicable for this release')

        Script._validate_type(asset_uid, u'asset_uid', STR_TYPE, True)
        Script._validate_type(rev_uid, u'rev_uid', int, False)

        params = self._client._params()

        if rev_uid is not None:
            params.update({'revision_id': rev_uid})

        import urllib
        if not self._ICP:
            asset_response = requests.get(self._client.service_instance._href_definitions.get_asset_href(asset_uid),
                                          params=params,
                                          headers=self._client._get_headers())
        else:
            asset_response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(asset_uid),
                                          params=params,
                                          headers=self._client._get_headers(), verify=False)
        asset_details = self._handle_response(200, u'get assets', asset_response)

        if self._WSD:
            attachment_url = asset_details['attachments'][0]['object_key']
            artifact_content_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                   urllib.parse.quote('script/' + attachment_url, safe='')

            r = requests.get(artifact_content_url, params=self._client._params(), headers=self._client._get_headers(),
                             stream=True, verify=False)
            if r.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading data asset"), r)

            downloaded_asset = r.content
            try:
                with open(filename, 'wb') as f:
                    f.write(downloaded_asset)
                print(u'Successfully saved data asset content to file: \'{}\''.format(filename))
                return os.getcwd() + "/" + filename
            except IOError as e:
                raise WMLClientError(u'Saving data asset with artifact_url: \'{}\'  to local file failed.'.format(filename), e)
        else:
            attachment_id = asset_details["attachments"][0]["id"]
            if not self._ICP:
                response = requests.get(self._client.service_instance._href_definitions.get_attachment_href(asset_uid,attachment_id), params=params,
                                        headers=self._client._get_headers())
            else:
                response = requests.get(self._client.service_instance._href_definitions.get_attachment_href(asset_uid,attachment_id), params=params,
                                          headers=self._client._get_headers(), verify=False)
            if response.status_code == 200:
                attachment_signed_url = response.json()["url"]
                if 'connection_id' in asset_details["attachments"][0]:
                    if not self._ICP:
                        att_response = requests.get(attachment_signed_url)
                    else:
                        att_response = requests.get(attachment_signed_url,
                                                    verify=False)
                else:
                    if not self._ICP:
                        att_response = requests.get(attachment_signed_url)
                    else:
                        att_response = requests.get(self._wml_credentials["url"]+attachment_signed_url,
                                                verify=False)
                if att_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading asset"), att_response)

                downloaded_asset = att_response.content
                try:
                    with open(filename, 'wb') as f:
                        f.write(downloaded_asset)
                    print(u'Successfully saved data asset content to file: \'{}\''.format(filename))
                    return os.getcwd() + "/" + filename
                except IOError as e:
                    raise WMLClientError(u'Saving asset with artifact_url to local file: \'{}\' failed.'.format(filename), e)
            else:
                raise WMLClientError("Failed while downloading the asset " + asset_uid)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(asset_details):
        """
                Get Unique Id of stored script asset.

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the stored script asset\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored shiny asset\n
                    **return type**: str\n

                **Example**

                     >>> asset_uid = client.script.get_id(asset_details)
        """

        return Script.get_uid(asset_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(asset_details):
        """
                Get Unique Id  of stored script asset. This method is deprecated. Use 'get_id(asset_details)' instead

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the stored script asset\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored script asset\n
                    **return type**: str\n

                **Example**

                 >>> asset_uid = client.script.get_uid(asset_details)
        """
        Script._validate_type(asset_details, u'asset_details', object, True)
        Script._validate_type_of_details(asset_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(asset_details, u'data_assets_details',
                                                           [u'metadata', u'guid'])


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(asset_details):
        """
            Get url of stored scripts asset.

           **Parameters**

           .. important::
                #. **asset_details**:  stored script details\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of stored script asset\n
                **return type**: str\n

           **Example**

             >>> asset_details = client.script.get_details(asset_uid)
             >>> asset_href = client.script.get_href(asset_details)

        """
        Script._validate_type(asset_details, u'asset_details', object, True)
        Script._validate_type_of_details(asset_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(asset_details, u'asset_details', [u'metadata', u'href'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def update(self, script_uid, meta_props=None, file_path=None):
        """
        Update script with either metadata or attachment or both.

        **Parameters**

        .. important::
            #. **param script_uid**:  Script UID.\n
               **type**: str\n

        **Output**

        .. important::

            **returns**: Updated metadata of script.\n
            **return type**: dict\n

        **Example**

            >>> script_details = client.script.update(model_uid, meta, content_path)
        """

        WMLResource._chk_and_block_create_update_for_python36(self)
        # We need to enable this once we add functionality for WSD
        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        Script._validate_type(script_uid, 'script_uid', STR_TYPE, True)

        if meta_props is None and file_path is None:
            raise WMLClientError('Atleast either meta_props or file_path has to be provided')

        updated_details = None
        details = None

        url = self._client.service_instance._href_definitions.get_asset_href(script_uid)

        # STEPS
        # STEP 1. Get existing metadata
        # STEP 2. If meta_props provided, we need to patch meta
        #   CAMS has meta and entity patching. 'name' and 'description' get stored in CAMS meta section
        #   a. Construct meta patch string and call /v2/assets/<asset_id> to patch meta
        #   b. Construct entity patch if required and call /v2/assets/<asset_id>/attributes/script to patch entity
        # STEP 3. If file_path provided, we need to patch the attachment
        #   a. If attachment already exists for the script, delete it
        #   b. POST call to get signed url for upload
        #   c. Upload to the signed url
        #   d. Mark upload complete
        # STEP 4. Get the updated script record and return

        # STEP 1
        if not self._ICP:
            response = requests.get(
                url,
                params=self._client._params(),
                headers=self._client._get_headers()
            )
        else:
            response = requests.get(
                url,
                params=self._client._params(),
                headers=self._client._get_headers(),
                verify=False
            )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of script_uid provided.')
            else:
                raise ApiRequestFailure(u'Failure during {}.'.format("getting script to update"), response)

        details = self._handle_response(200, "Get script details", response)

        attachments_response = None

        # STEP 2a.
        # Patch meta if provided
        if meta_props is not None:
            self._validate_type(meta_props, u'meta_props', dict, True)
            meta_props_str_conv(meta_props)

            meta_patch_payload = []
            entity_patch_payload = []

            # Since we are dealing with direct asset apis, there can be metadata or entity patch or both
            if "name" in meta_props or "description" in meta_props:
                props_for_asset_meta_patch = {}

                for key in meta_props:
                    if key == 'name' or key == 'description':
                        props_for_asset_meta_patch.update({key: meta_props[key]})

                meta_patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details,
                                                                                         props_for_asset_meta_patch,
                                                                                         with_validation=True)
            # STEP 2b.
            if "software_spec_uid" in meta_props:
                if details[u'entity'][u'script'][u'software_spec']:
                    entity_patch_payload = [{'op':'replace',
                                             'path': '/software_spec/base_id',
                                             'value': meta_props[u'software_spec_uid']}]
                else:
                    entity_patch_payload = [{'op': 'add',
                                             'path': '/software_spec',
                                             'value': '{base_id:' + meta_props[u'software_spec_uid'] + '}'}]

            if meta_patch_payload:
                meta_patch_url = self._client.service_instance._href_definitions.get_asset_href(script_uid)

                if not self._ICP:
                    response_patch = requests.patch(meta_patch_url,
                                                    json=meta_patch_payload,
                                                    params=self._client._params(),
                                                    headers=self._client._get_headers())
                else:
                    response_patch = requests.patch(meta_patch_url,
                                                    json=meta_patch_payload,
                                                    params=self._client._params(),
                                                    headers=self._client._get_headers(),verify=False)

                updated_details = self._handle_response(200, u'script patch', response_patch)

            if entity_patch_payload:
                entity_patch_url = self._client.service_instance._href_definitions.get_asset_href(script_uid) + '/attributes/script'

                if not self._ICP:
                    response_patch = requests.patch(entity_patch_url,
                                                    json=entity_patch_payload,
                                                    params=self._client._params(),
                                                    headers=self._client._get_headers())
                else:
                    response_patch = requests.patch(entity_patch_url,
                                                    json=entity_patch_payload,
                                                    params=self._client._params(),
                                                    headers=self._client._get_headers(), verify=False)

                updated_details = self._handle_response(200, u'script patch', response_patch)

        if file_path is not None:
            if "attachments" in details and details[u'attachments']:
                current_attachment_id = details[u'attachments'][0][u'id']
            else:
                current_attachment_id = None

            #STEP 3
            attachments_response = self._update_attachment_for_assets("script",
                                                                      script_uid,
                                                                      file_path,
                                                                      current_attachment_id)

        if attachments_response is not None and 'success' not in attachments_response:
            self._update_msg(updated_details)

        # Have to fetch again to reflect updated asset and attachment ids
        url = self._client.service_instance._href_definitions.get_asset_href(script_uid)

        if not self._ICP:
            response = requests.get(
                url,
                params=self._client._params(),
                headers=self._client._get_headers()
            )
        else:
            response = requests.get(
                url,
                params=self._client._params(),
                headers=self._client._get_headers(),
                verify=False
            )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of script_uid provided.')
            else:
                raise ApiRequestFailure(u'Failure during {}.'.format("getting script to update"), response)

        # response = self._handle_response(200, "Get script details", response)

        # return self._get_required_element_from_response(response)

        response = self._get_required_element_from_response(self._handle_response(200, "Get script details", response))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            entity = response[u'entity']

            try:
                del entity[u'script'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

    def _update_msg(self, updated_details):
        if updated_details is not None:
            print("Could not update the attachment because of server error."
                  " However metadata is updated. Try updating attachment again later")
        else:
            raise WMLClientError('Unable to update attachment because of server error. Try again later')

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, asset_uid):
        """
            Delete a stored script asset.

            **Parameters**

            .. important::
                #. **asset_uid**:  Unique Id of script asset\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.script.delete(asset_uid)

        """
        Script._validate_type(asset_uid, u'asset_uid', STR_TYPE, True)
        if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and \
                self._if_deployment_exist_for_asset(asset_uid):
            raise WMLClientError(
                u'Cannot delete script that has existing deployments. Please delete all associated deployments and try again')

        if not self._ICP:
            response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                    headers=self._client._get_headers())
        else:
            response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                      headers=self._client._get_headers(), verify=False)
        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete assets', response)


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_revision(self, script_uid):
        """
        Creates revision for the given script. Revisions are immutable once created.
        The metadata and attachment at script_uid is taken and a revision is created out of it

        **Parameters**

        .. important::
            #. **script_uid**: Script ID. Mandatory.\n
               **type**: str\n

        ** Output**

        .. important::

            **returns**: Stored script revisions metadata.\n
            **return type**: dict\n

        **Example**

            >>> script_revision = client.scripts.create_revision(script_uid)
        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        # We need to enable this once we add functionality for WSD
        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()
        script_uid = str_type_conv(script_uid)
        Script._validate_type(script_uid, u'script_uid', STR_TYPE, True)

        print("Creating script revision...")

        # return self._get_required_element_from_response(
        #     self._create_revision_artifact_for_assets(script_uid, 'Script'))

        response = self._get_required_element_from_response(
            self._create_revision_artifact_for_assets(script_uid, 'Script'))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            entity = response[u'entity']

            try:
                del entity[u'script'][u'ml_version']
            except KeyError:
                pass

            final_response = {
                "metadata": response[u'metadata'],
                "entity": entity
            }

            return final_response

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def list_revisions(self, script_uid, limit=None):
        """
        List all revisions for the given script uid.

        **Parameters**

        .. important::
            #. **script_uid**: Stored script ID.\n
               **type**: str\n

            #. **limit**: Limit number of fetched records (optional).\n
               **type**: int\n

        **Output**

        .. important::
                This method only prints the list of all script in a table format.\n
                **return type**: None\n

        **Example**

            >>> client.scripts.list_revisions(script_uid)
        """
        # We need to enable this once we add functionality for WSD
        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()

        script_uid = str_type_conv(script_uid)
        Script._validate_type(script_uid, u'script_uid', STR_TYPE, True)

        url = self._client.service_instance._href_definitions.get_asset_href(script_uid) + "/revisions"
        # /v2/assets/{asset_id}/revisions returns 'results' object
        script_resources = self._get_with_or_without_limit(url,
                                                           limit,
                                                           'List Script revisions',
                                                           summary=None,
                                                           pre_defined=None)[u'resources']
        script_values = [
            (m[u'metadata'][u'asset_id'],
             m[u'metadata'][u'revision_id'],
             m[u'metadata'][u'name'],
             m[u'metadata'][u'commit_info'][u'committed_at']) for m in
            script_resources]

        self._list(script_values, [u'GUID', u'REVISION_ID', u'NAME', u'REVISION_COMMIT'], limit, _DEFAULT_LIST_LENGTH)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_revision_details(self, script_uid=None, rev_uid=None):
        """
        Get metadata of script_uid revision.

        **Paramaters**

        .. important::
            #. **script_uid**: Script ID. Mandatory.\n
               **type**: str\n

            #. **rev_uid**: Revision ID. If this parameter is not provided, returns latest revision if existing else error.\n
               **type**: int\n

        **Output**

        .. important::

            **returns**: Stored script(s) metadata\n
            **return type**: dict\n

        **Example**

            >>> script_details = client.scripts.get_revision_details(script_uid, rev_uid)
        """
        # We need to enable this once we add functionality for WSD
        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()

        script_uid = str_type_conv(script_uid)
        Script._validate_type(script_uid, u'script_uid', STR_TYPE, True)
        Script._validate_type(rev_uid, u'rev_uid', int, False)

        if rev_uid is None:
            rev_uid = 'latest'

        url = self._client.service_instance._href_definitions.get_asset_href(script_uid)
        # return self._get_required_element_from_response(self._get_with_or_without_limit(url,
        #                                        limit=None,
        #                                        op_name="asset_revision",
        #                                        summary=None,
        #                                        pre_defined=None,
        #                                        revision=rev_uid))
        resources = self._get_with_or_without_limit(url,
                                                    limit=None,
                                                    op_name="asset_revision",
                                                    summary=None,
                                                    pre_defined=None,
                                                    revision=rev_uid)['resources']
        responses = [self._get_required_element_from_response(resource) for resource in resources]

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return responses
        else:
            final_responses = []
            for response in responses:
                entity = response[u'entity']

                try:
                    del entity[u'script'][u'ml_version']
                except KeyError:
                    pass

                final_responses.append({
                    "metadata": response[u'metadata'],
                    "entity": entity
                })

            return final_responses

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def _delete(self, asset_uid):
        Script._validate_type(asset_uid, u'asset_uid', STR_TYPE, True)

        if not self._ICP:
            response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                       headers=self._client._get_headers())
        else:
            response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                       headers=self._client._get_headers(), verify=False)

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'scripts', dict)

        revision_id = None

        try:
            if self._client.default_space_id is not None:
                metadata = {'space_id': response_data['metadata']['space_id'],
                            'name':response_data['metadata']['name'],
                            'guid': response_data['metadata']['asset_id'],
                            'href': response_data['href'],
                            'asset_type': response_data['metadata']['asset_type'],
                            'created_at': response_data['metadata']['created_at'],
                            'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                            }
                if 'description' in response_data[u'metadata']:
                    metadata.update(
                        {'description':response_data[u'metadata'][u'description']})

                if self._client.ICP_30 is not None or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    if "revision_id" in response_data[u'metadata']:
                        revision_id = response_data[u'metadata'][u'revision_id']
                        metadata.update({'revision_id': response_data[u'metadata'][u'revision_id']})

                    if "attachments" in response_data and response_data[u'attachments']:
                        metadata.update({'attachment_id': response_data[u'attachments'][0][u'id']})

                    if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                        metadata.update(
                            {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                new_el = {'metadata': metadata,
                          'entity': response_data['entity']
                          }
            elif self._client.default_project_id is not None:
                if self._client.WSD:

                    href = self._client.service_instance._href_definitions.get_base_asset_href(response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

                    metadata = {'project_id': response_data['metadata']['project_id'],
                                'guid': response_data['metadata']['asset_id'],
                                'name': response_data['metadata']['name'],
                                'href': href,
                                'asset_type': response_data['metadata']['asset_type'],
                                'created_at': response_data['metadata']['created_at']
                                }

                    if 'description' in response_data[u'metadata']:
                        metadata.update(
                            {'description': response_data[u'metadata'][u'description']})

                    new_el = {'metadata': metadata,
                              'entity': response_data['entity']
                              }
                    if self._client.WSD_20 is not None:
                        if "revision_id" in response_data[u'metadata']:
                            revision_id = response_data[u'metadata'][u'revision_id']
                            new_el['metadata'].update({'revision_id': response_data[u'metadata'][u'revision_id']})

                        if "attachments" in response_data and response_data[u'attachments']:
                            new_el['metadata'].update({'attachment_id': response_data[u'attachments'][0][u'id']})

                        if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                            new_el['metadata'].update(
                                {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                    if 'usage' in response_data['metadata']:
                        new_el['metadata'].update({'last_updated_at': response_data['metadata']['usage']['last_updated_at']})
                    else:
                        new_el['metadata'].update(
                         {'last_updated_at': response_data['metadata']['last_updated_at']})
                else:
                    metadata = {'project_id': response_data['metadata']['project_id'],
                                'guid': response_data['metadata']['asset_id'],
                                'href': response_data['href'],
                                'name': response_data['metadata']['name'],
                                'asset_type': response_data['metadata']['asset_type'],
                                'created_at': response_data['metadata']['created_at'],
                                'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                                }
                    if self._client.ICP_30 is not None or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                        if "revision_id" in response_data[u'metadata']:
                            revision_id = response_data[u'metadata'][u'revision_id']
                            metadata.update({'revision_id': response_data[u'metadata'][u'revision_id']})

                        if "attachments" in response_data and response_data[u'attachments']:
                            metadata.update({'attachment_id': response_data[u'attachments'][0][u'id']})

                        if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                            metadata.update(
                                {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                    if 'description' in response_data[u'metadata']:
                        metadata.update(
                            {'description': response_data[u'metadata'][u'description']})
                    new_el = {'metadata': metadata,
                              'entity': response_data['entity']
                              }
                    if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                        href_without_host = response_data['href'].split('.com')[-1]
                        new_el[u'metadata'].update({'href': href_without_host})
            return new_el
        except Exception as e:
            raise WMLClientError(f"Failed to read Response from down-stream service: {response_data}")



# coding: utf-8

"""
Copyright 2016 SmartBear Software

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Ref: https://github.com/swagger-api/swagger-codegen
"""

#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems


class FrameworkOutputRepository(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, version=None ,runtime= None, runtimes = None, libraries = None):
        """
        FrameworkOutputRepository - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'version': 'str',
            'runtime' : 'FrameworkOutputRepositoryRuntimes',
            'runtimes': 'list[FrameworkOutputRepositoryRuntimes]',
            'libraries': 'list[FrameworkOutputRepositoryLibraries]'
        }

        self.attribute_map = {
            'name': 'name',
            'version': 'version',
            'runtime' : 'runtime',
            'runtimes': 'runtimes',
            'libraries': 'libraries'
        }

        self._name = name
        self._version = version
        self._runtime = runtime
        self._runtimes = runtimes
        self._libraries = libraries

    @property
    def name(self):
        """
        Gets the name of this FrameworkOutputRepository.


        :return: The name of this FrameworkOutputRepository.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this FrameworkOutputRepository.


        :param name: The name of this FrameworkOutputRepository.
        :type: str
        """
        self._name = name

    @property
    def version(self):
        """
        Gets the version of this FrameworkOutputRepository.


        :return: The version of this FrameworkOutputRepository.
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """
        Sets the version of this FrameworkOutputRepository.


        :param version: The version of this FrameworkOutputRepository.
        :type: str
        """
        self._version = version

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result


    @property
    def runtimes(self):
        """
        Gets the runtimes of this FrameworkOutputRepository.


        :return: The runtimes of this FrameworkOutputRepository.
        :rtype: list[FrameworkOutputRepositoryRuntimes]
        """
        return self._runtimes

    @runtimes.setter
    def runtimes(self, runtimes):
        """
        Sets the runtimes of this FrameworkOutputRepository.


        :param runtimes: The runtimes of this FrameworkOutputRepository.
        :type: list[FrameworkOutputRepositoryRuntimes]
        """
        self._runtimes = runtimes

    @property
    def libraries(self):
        """
        Gets the libraries of this FrameworkOutputRepository.


        :return: The libraries of this FrameworkOutputRepository.
        :rtype: list[FrameworkOutputRepositoryLibraries]
        """

        return self._libraries

    @libraries.setter
    def libraries(self, libraries):
        """
        Sets the libraries of this FrameworkOutputRepository.

        :param runtimes: The runtimes of this FrameworkOutputRepository.
        :type: list[FrameworkOutputRepositoryLibraries]
        """
        self._libraries = libraries


    @property
    def runtime(self):
        """
        Gets the runtime of this FrameworkOutputRepository.


        :return: The runtimes of this FrameworkOutputRepository.
        :rtype: list[FrameworkOutputRepositoryRuntimes]
        """
        return self._runtime

    @runtime.setter
    def runtime(self, runtime):
        """
        Sets the runtimes of this FrameworkOutputRepository.


        :param runtimes: The runtimes of this FrameworkOutputRepository.
        :type: list[FrameworkOutputRepositoryRuntimes]
        """
        self._runtimes = runtime

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




import joblib
import numpy as np
import logging

from diffprivlib.models import GaussianNB

from ibmfl.exceptions import ModelInitializationException, ModelException
from ibmfl.model.model_update import ModelUpdate
from ibmfl.model.sklearn_fl_model import SklearnFLModel

logger = logging.getLogger(__name__)


class NaiveBayesFLModel(SklearnFLModel):
    """
    Wrapper class for diffprivlib.models.GaussianNB classifier, which itself
    is an implementation of sklearn.naive_bayes.GaussianNB with differential
    privacy.
    """

    def __init__(self, model_name, model_spec, model=None, **kwargs):
        """
        Create a `NaiveBayesFLModel` instance from a diffprivlib.models.GaussianNB
        model. If model is provided, it will use it; otherwise it will take
        the model_spec to create the model.

        :param model_name: A name specifying the type of model, e.g., \
        naive_bayes
        :type model_name: `str`
        :param model_spec: A dictionary contains model specification
        :type model_spec: `dict`
        :param model: Compiled GaussianNB model
        :type model: `diffprivlib.models.GaussianNB`
        :param kwargs: A dictionary contains other parameter settings on \
         to initialize a Naive Bayes model.
        :type kwargs: `dict`
        """
        if model:
            if not issubclass(type(model), GaussianNB):
                raise ValueError('Compiled GaussianNB needs to be provided'
                                 '(diffprivlib.models.GaussianNB). '
                                 'Type provided' + str(type(model)))

        super().__init__(model_name,
                         model_spec,
                         sklearn_model=model,
                         **kwargs)

        self.old_vals = {"theta": None, "sigma": None, "class_count": None}
        self.model_trained = False
        self.model_type = 'Sklearn-NaiveBayes'

    def fit_model(self, train_data, fit_params=None, **kwargs):
        """
        Fits current model with provided training data.

        :param train_data: Training data a tuple given in the form \
        (x_train, y_train).
        :type train_data: `np.ndarray`
        :param fit_params: Not used, present here for API consistency by \
        convention.
        :return: None
        """
        self.model.fit(train_data[0], train_data[1])
        self.model_trained = True

    def update_model(self, model_update, **kwargs):
        """
        Update GaussianNB model with provided model_update.

        :param model_update: `ModelUpdate` object that contains \
        the required information to update the GaussianNB model.
        :type model_update: `ModelUpdate`
        :return: None
        """
        if model_update.get("class_count") is None:
            return

        self.model.theta_ = model_update.get("theta")
        self.model.sigma_ = model_update.get("sigma")
        self.model.class_count_ = model_update.get("class_count")

        self._store_old_values(self.model.theta_, self.model.sigma_,
                               self.model.class_count_)

    def get_model_update(self):
        """
        Generates a `ModelUpdate` object that will be sent to other entities.

        :return: ModelUpdate
        :rtype: `ModelUpdate`
        """
        if (not self.model_trained or
                not hasattr(self.model, "class_count_") or
                np.all(self.old_vals['class_count'] ==
                       self.model.class_count_)):
            return ModelUpdate(theta=None, sigma=None, class_count=None)
        if self.old_vals['class_count'] is None:
            return ModelUpdate(theta=self.model.theta_,
                               sigma=self.model.sigma_,
                               class_count=self.model.class_count_)

        class_count_update = self.model.class_count_ - \
            self.old_vals['class_count']

        nm = self.model.class_count_[:, np.newaxis]
        n = self.old_vals['class_count'][:, np.newaxis]
        m = class_count_update[:, np.newaxis]

        theta_update = self.model.theta_ * nm - self.old_vals['theta'] * n
        theta_update = theta_update / m

        sigma_update = nm * self.model.sigma_ - n * self.old_vals['sigma'] - \
            (n * m) / nm * (self.old_vals['theta'] - theta_update) ** 2

        return ModelUpdate(theta=theta_update,
                           sigma=sigma_update,
                           class_count=class_count_update)

    def evaluate(self, test_dataset, **kwargs):
        """
        Evaluates the model given testing data.
        :param test_dataset: Testing data, a tuple given in the form \
        (x_test, test) or a datagenerator of of type `keras.utils.Sequence`, 
        `keras.preprocessing.image.ImageDataGenerator`
        :type test_dataset: `np.ndarray`

        :param kwargs: Dictionary of metrics available for the model
        :type kwargs: `dict`
        """

        if type(test_dataset) is tuple:
            x_test = test_dataset[0]
            y_test = test_dataset[1]

            return self.evaluate_model(x_test, y_test)

        else:
            raise ModelException("Invalid test dataset!")

    def evaluate_model(self, x, y, **kwargs):
        """
        Evaluates the model given test data x and the corresponding labels y.

        :param x: Samples with shape as expected by the model.
        :type x: `np.ndarray`
        :param y: Corresponding true labels to x
        :type y: `np.ndarray`
        :param kwargs: Dictionary of model-specific arguments for evaluating \
        models. For example, sample weights accepted by model.score.
        :return: Dictionary with all evaluation metrics provided by \
        specific implementation.
        :rtype: `dict`
        """
        acc = {'score': self.model.score(x, y, **kwargs)}
        return acc

    def _store_old_values(self, theta, sigma, class_count):
        """
        Store old values of training parameters to allow reconstruction of
        parameters for newly-trained examples when model update is sought.
        """
        self.old_vals = {"theta": theta,
                         "sigma": sigma,
                         "class_count": class_count}

    @staticmethod
    def load_model_from_spec(model_spec):
        """
        Loads model from provided model_spec, where model_spec is a `dict`
        that contains the following items: model_spec['model_definition']
        contains the model definition as type diffprivlib.GaussianNB.

        :param model_spec: Model specification contains \
        a complied sklearn model.
        :param model_spec: `dict`
        :return: model
        :rtype: `sklearn.cluster`
        """
        model = None
        try:
            if 'model_definition' in model_spec:
                path = model_spec['model_definition']
                with open(path, 'rb') as f:
                    model = joblib.load(f)

                if not issubclass(type(model), GaussianNB):
                    raise ValueError('Provided complied model in model_spec '
                                     'should be of type GaussianNB.'
                                     'Instead they are:' + str(type(model)))
        except Exception as ex:
            raise ModelInitializationException('Model specification was '
                                               'badly formed. ', str(ex))
        return model



import unittest
import pandas as pd

from ibm_watson_machine_learning import APIClient
from ibm_watson_machine_learning.experiment import AutoAI
from ibm_watson_machine_learning.deployment import WebService
from ibm_watson_machine_learning.tests.utils import get_wml_credentials, get_cos_credentials, is_cp4d, bucket_exists, bucket_name_gen
from ibm_watson_machine_learning.experiment.autoai.optimizers import LocalAutoPipelines

from sklearn.pipeline import Pipeline


class TestAutoAILocal(unittest.TestCase):
    wml_client: 'APIClient' = None
    experiment: 'AutoAI' = None
    local_optimizer = None
    data_location = './autoai/data/bank.csv'
    target_column = 'y'
    train_data = None
    test_X = None
    test_y = None
    pipeline_name = 'Pipeline_3'
    OPTIMIZER_NAME = 'Local ai4ml'

    @classmethod
    def setUp(cls) -> None:
        cls.experiment = AutoAI()
        cls.data = pd.read_csv(cls.data_location)
        cls.X = cls.data.drop([cls.target_column], axis=1)
        cls.y = cls.data[cls.target_column]

    def test_01_initialize_optimizer(self):
        TestAutoAILocal.local_optimizer = self.experiment.optimizer(
            name=self.OPTIMIZER_NAME,
            prediction_type=AutoAI.PredictionType.MULTICLASS,
            prediction_column=self.target_column,
            scoring=AutoAI.Metrics.ACCURACY_SCORE,
            _force_local_scenario=True
        )

        self.assertIsInstance(self.local_optimizer, LocalAutoPipelines,
                              msg="experiment.optimizer did not return RemoteAutoPipelines object")

    def test_02_get_configuration_parameters_of_remote_auto_pipeline(self):
        parameters = self.local_optimizer.get_params()
        print(parameters)
        self.assertIsInstance(parameters, dict, msg='Config parameters are not a dictionary instance.')

    def test_03_fit_optimizer(self):
        best_model = self.local_optimizer.fit(X=TestAutoAILocal.X, y=TestAutoAILocal.y)
        self.assertIsInstance(best_model, Pipeline)

    def test_04_get_pipeline_model_details(self):
        details = self.local_optimizer.get_pipeline_details(pipeline_name='Pipeline_2')
        self.assertIn('pipeline_nodes', details)

    def test_05_get_holdout_data(self):
        TestAutoAILocal.test_X, TestAutoAILocal.test_y = self.local_optimizer.get_holdout_data()

        print("holdout data sample:")
        print(self.test_X.head())
        self.assertGreater(len(self.test_X), 0)

    def test_06_get_pipeline_model(self):
        pipeline_model = self.local_optimizer.get_pipeline(pipeline_name=TestAutoAILocal.pipeline_name, astype='sklearn')
        results = pipeline_model.predict(X=TestAutoAILocal.test_X.values)

        self.assertIsInstance(pipeline_model, Pipeline)
        print(results)
        self.assertGreater(len(results), 0)

    def test_06_predict_with_optimizer(self):
        results = self.local_optimizer.predict(X=TestAutoAILocal.test_X.values)
        self.assertGreater(len(results), 0)


    #################################
    #      DEPLOYMENT SECTION       #
    #################################

    # def test_21_deployment_setup_and_preparation(self):
    #     if is_cp4d():
    #         TestAutoAILocal.service = WebService(wml_credentials=self.wml_credentials.copy(),
    #                                              project_id=self.project_id,
    #                                              space_id=self.space_id)
    #         self.wml_client.set.default_space(self.space_id)
    #     else:
    #         TestAutoAILocal.service = WebService(wml_credentials=self.wml_credentials)
    #
    # def test_22__deploy__deploy_best_computed_pipeline_from_autoai_on_wml(self):
    #     self.service.create(
    #         experiment_run_id=self.run_id,
    #         model=self.new_pipeline,
    #         deployment_name=self.DEPLOYMENT_NAME)
    #
    # def test_24_score_deployed_model(self):
    #     nb_records = 10
    #     predictions = self.service.score(payload=self.holdout_data.drop([self.target_column], axis=1)[:nb_records])
    #     print(predictions)
    #     self.assertIsNotNone(predictions)
    #     self.assertEqual(len(predictions['predictions'][0]['values']), nb_records)
    #
    #
    # def test_25_delete_deployment(self):
    #     print("Delete current deployment: {}".format(self.service.deployment_id))
    #     self.service.delete()

if __name__ == '__main__':
    unittest.main()



import unittest
from preparation_and_cleaning import *
import logging
from preparation_and_cleaning import *


class TestWMLClientWithSpace(unittest.TestCase):
    space_uid = None
    space_href=None
    member_uid = None
    member_href = None

    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithSpace.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()


    def test_01_service_instance_details(self):
        TestWMLClientWithSpace.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestWMLClientWithSpace.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()
        TestWMLClientWithSpace.logger.debug(details)

        self.assertIsInstance(details, dict)

    def test_02_save_space(self):
        metadata = {
            self.client.spaces.ConfigurationMetaNames.NAME: "V4Space"
        }

        space_details = self.client.spaces.store(meta_props=metadata)

        TestWMLClientWithSpace.space_uid = self.client.spaces.get_uid(space_details)

        space_specific_details = self.client.spaces.get_details(TestWMLClientWithSpace.space_uid)
        self.assertTrue(TestWMLClientWithSpace.space_uid in str(space_specific_details))

    def test_03_update_space(self):
        metadata = {
            self.client.spaces.ConfigurationMetaNames.NAME: "my_space",
        }

        space_details = self.client.spaces.update(TestWMLClientWithSpace.space_uid, metadata)
        self.assertTrue('my_space' in str(space_details))
        TestWMLClientWithSpace.logger.info(space_details)
        self.assertTrue('V4Space' not in str(space_details))

    def test_04_get_space_details(self):
        details = self.client.spaces.get_details()
        self.assertTrue(TestWMLClientWithSpace.space_uid in str(details))

        details2 = self.client.spaces.get_details(TestWMLClientWithSpace.space_uid)
        self.assertTrue(TestWMLClientWithSpace.space_uid in str(details2))

    def test_05_list(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.spaces.list() # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithSpace.space_uid in captured_output.getvalue())
        self.client.spaces.list()  # Just to see values.

    def test_06_create_member(self):
        metadata = {
            self.client.spaces.MemberMetaNames.MEMBERS: [{
                "id": "17",
                "type": "user",
                "role": "viewer"
            }]
        }

        member_details = self.client.spaces.create_member(
            space_id=TestWMLClientWithSpace.space_uid,
            meta_props=metadata
        )
        TestWMLClientWithSpace.member_uid = self.client.spaces.get_member_uid(member_details)
        TestWMLClientWithSpace.member_href = self.client.spaces.get_member_href(member_details)

        member_specific_details = self.client.spaces.get_member_details(
            TestWMLClientWithSpace.space_uid,
            TestWMLClientWithSpace.member_uid
        )
        self.assertTrue(TestWMLClientWithSpace.member_uid in str(member_specific_details))

    def test_07_get_member_details(self):
        member_specific_details = self.client.spaces.get_member_details(
            TestWMLClientWithSpace.space_uid,
            TestWMLClientWithSpace.member_uid
        )
        self.assertTrue(TestWMLClientWithSpace.member_uid in str(member_specific_details))

    def test_09_delete_member(self):
        self.client.spaces.delete_member(
            TestWMLClientWithSpace.space_uid,
            TestWMLClientWithSpace.member_uid
        )

    def test_10_delete_space(self):
        self.client.spaces.delete(TestWMLClientWithSpace.space_uid)



import unittest

import logging
from ibm_watson_machine_learning.tests.CP4D_35.preparation_and_cleaning import *

class TestHwSpec(unittest.TestCase):

    logger = logging.getLogger(__name__)

    @classmethod
    def setUpClass(self):
        TestHwSpec.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

    def test_01_list_hw_specs(self):

        hw_spec_details1 = self.client.hardware_specifications.list()
        print(hw_spec_details1)

        hw_spec_details2 = self.client.hardware_specifications.list(name='V100x2')
        print(hw_spec_details2)

    def test_02_get_details(self):

        hw_spec_id = self.client.hardware_specifications.get_uid_by_name('V100x2')

        details = self.client.hardware_specifications.get_details(hw_spec_id)
        print(details)
        self.assertTrue("V100x2" in str(details))

if __name__ == '__main__':
    unittest.main()



# coding: utf-8

"""
    IBM Watson Machine Learning REST API

    ## Authorization  ### IBM Watson Machine Learning Credentials  To start working with API one needs to generate an `access token` using the `username` and `password` available on the Service Credentials tab of the IBM Watson Machine Learning service instance or also available in the VCAP environment variable.  Example of the Service Credentials:  ```json {     \"url\": \"https://ibm-watson-ml.mybluemix.net\",     \"access_key\": \"ERY9vcBfE4sE+F4g8hcotF9L+j81WXWeZv\",     \"username\": \"c1ef4b80-2ee2-458e-ab92-e9ca97ec657d\",     \"password\": \"030528d4-5a3e-4d4c-9258-5d553513be6f\" } ```  Example of obtaining `access token` from Token Endpoint using HTTP Basic Auth (for details please refer to Token section below):  ` curl --basic --user username:password https://ibm-watson-ml.mybluemix.net/v2/identity/token `   ### Apache Spark Service Credentials  The IBM Watson Machine Learning co-operates with the Apache Spark as a Service to deploy pipeline models. For API methods requiring Apache Spark Service instance a custom header: `X-Spark-Service-Instance` with Service Credentials must be specified. The header value is a **base64 encoded** string with the Service Credentials JSON data.  [Example of API method requiring Apache Spark Service](https://console.ng.bluemix.net/docs/services/PredictiveModeling/index-gentopic1.html#pm_service_api_spark_batch) 

    OpenAPI spec version: 2.0.0
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

#  (C) Copyright IBM Corp. 2020.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#       http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pprint import pformat
from six import iteritems
import re


class EvaluationDefinition(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, method=None, metrics=None):
        """
        EvaluationDefinition - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'method': 'str',
            'metrics': 'list[EvaluationDefinitionMetrics]'
        }

        self.attribute_map = {
            'method': 'method',
            'metrics': 'metrics'
        }

        self._method = method
        self._metrics = metrics

    @property
    def method(self):
        """
        Gets the method of this EvaluationDefinition.
        The name of the evaluation algorithm

        :return: The method of this EvaluationDefinition.
        :rtype: str
        """
        return self._method

    @method.setter
    def method(self, method):
        """
        Sets the method of this EvaluationDefinition.
        The name of the evaluation algorithm

        :param method: The method of this EvaluationDefinition.
        :type: str
        """

        self._method = method

    @property
    def metrics(self):
        """
        Gets the metrics of this EvaluationDefinition.


        :return: The metrics of this EvaluationDefinition.
        :rtype: list[EvaluationDefinitionMetrics]
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """
        Sets the metrics of this EvaluationDefinition.


        :param metrics: The metrics of this EvaluationDefinition.
        :type: list[EvaluationDefinitionMetrics]
        """

        self._metrics = metrics

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
"""
Serialization factory provides a way to create a serializer
and deserializer to convert byte streams to a message and
vice versa
"""
from ibmfl.message.json_serializer import JSONSerializer
from ibmfl.message.pickle_serializer import PickleSerializer
from ibmfl.message.serializer_types import SerializerTypes


class SerializerFactory(object):
    """
    Class for a factory to serialize and deserialize
    """
    def __init__(self, serializer_type):
        """
        Creates an object of `SerializerFactory` class

        :param serializer_type: type of seriaze and deserialize
        :type serializer_type: `Enum`
        """
        self.serializer = None
        if serializer_type is SerializerTypes.PICKLE:
            self.serializer = PickleSerializer()
        elif serializer_type is SerializerTypes.JSON_PICKLE:
            self.serializer = JSONSerializer()

    def build(self):
        """
        Returns a serializer

        :param: None
        :return: serializer
        :rtype: `Serializer`
        """
        return self.serializer



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
"""
Create routes for the server handler
"""
import logging
import json
from ibmfl.message.message_type import MessageType
from ibmfl.message.message import Message

logger = logging.getLogger(__name__)


def default_end_point(message):
    """
    All Requests whose request type is not determined are routed to this method
    Primary function is to log the request received
    :param request: request received by the server
    :type request: `Message`
    :param response: response to send back
    :type response: `Message`
    :param router: router associated with the request
    :type router: `Router`
    :param kwargs: Dictionary of model-specific arguments.
    :type kwargs: `dict`

    """
    # more detailed logging should be done on request message

    logging.info(message.__dict__)

    message.set_data({'status': 'error'})

    return message


def get_aggregator_router(router, agg_proto_handler):
    """
    Route for register party

    :param router: Router object
    :type router: `Router`
    :param agg_proto_handler: ProtoHandler object
    :type agg_proto_handler: `ProtoHandler`
    :return: None
    """
    router.add_routes({
        '{}'.format(MessageType.REGISTER.value):
        agg_proto_handler.register_party,

        '{}'.format(MessageType.TRAIN.value):
        agg_proto_handler.process_model_update_requests,

        'default': default_end_point
    })


def get_party_router(router, party_proto_handler):
    """
    Route for register party

    :param router: Router object
    :type router: `Router`
    :param party_proto_handler: ProtoHandler object
    :type party_proto_handler: `ProtoHandler`
    :return: None
    """
    router.add_routes({
        '{}'.format(MessageType.TRAIN.value): party_proto_handler.handle_async_request,
        '{}'.format(MessageType.SAVE_MODEL.value): party_proto_handler.handle_request,
        '{}'.format(MessageType.EVAL_MODEL.value): party_proto_handler.handle_request,
        '{}'.format(MessageType.SYNC_MODEL.value): party_proto_handler.handle_request,
        '{}'.format(MessageType.STOP.value): party_proto_handler.handle_request,
        'default': default_end_point
    })



"""
Module acting as a bridge between Server and ProtocolHandler
Routes for all the requests received by the server are redirected
to designated handler in PH using routers
"""
import logging
from parse import parse
from ibmfl.exceptions import DuplicateRouteException


logger = logging.getLogger(__name__)


class Router(object):
    """
    Container used to add and match a group of routes.
    """

    def __init__(self):
        self.routes = {}

    def add_route(self, path, handler):
        """
        Add path, handler combination in a dictionary.
        Throws DuplicateRouteException when route is repeated

        :param path: special pre defined /assigned identifier which will be \
            used by the client to reach specific handler in the PH
        :param handler: method which needs to handle the request \
            carrying the special identifier
        :raises `DuplicateRouteException`: exception is raised \
            when path is added multiple times
        """
        if path not in self.routes:
            self.routes[path] = handler
        else:
            raise DuplicateRouteException

    def add_routes(self, routes):
        """
        Add multiple path, handler combinations in routes dictionary.
        Throws DuplicateRouteException when route is repeated

        :param routes: list of path, handler combination
        :raises `DuplicateRouteException`: exception is raised when \
        path is added multiple times
        """
        for route, handler in routes.items():
            self.add_route(route, handler)

    def get_handler(self, request_path):
        """
        Gets handler for path specified in the request

        :param request_path: path or identifier provided in \
            the request message
        :return: a handler which was assigned when Router object \
            was created
        """
        for path, handler in self.routes.items():
            parse_result = parse(path, request_path)
            if parse_result is not None:
                return handler, parse_result.named

        return None, None



#
#    IBM Confidential
#    OCO Source Materials
#    [5737-H76, 5725-W78, 5900-A1R, 5737-L65]
#    (C) Copyright IBM Corp. 2020  All Rights Reserved.
#    The source code for this program is not published or otherwise divested of its trade secrets,
#    irrespective of what has been deposited with the U.S. Copyright Office.


class GlobalizationUtil:

    @staticmethod
    def get_language():
        language = "en"
        return language



__all__ = [
    'DataSourceTypeNotRecognized',
    'LabelNotFound',
    'APIConnectionError',
    'DataStreamError',
    'WrongDatabaseSchemaOrTable',
    'WrongFileLocation',
    'WrongLocationProperty',
    'UnsupportedConnection',
    'UnsupportedOutputConnection',
    'MissingFileName',
    'FileUploadFailed',
    'FlightServiceInfoNotAvailable'
]

from ibm_watson_machine_learning.wml_client_error import WMLClientError


class DataSourceTypeNotRecognized(WMLClientError, NotImplementedError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Data source type: {value_name} not recognized!", reason)


class LabelNotFound(WMLClientError, KeyError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Cannot find label: {value_name}", reason)


class APIConnectionError(WMLClientError, ConnectionError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"Cannot connect to: {value_name}", reason)


class DataStreamError(WMLClientError, ConnectionError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. Try again.", reason)


class WrongLocationProperty(WMLClientError, ConnectionError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. Try again.", reason)


class WrongFileLocation(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. Try again.", reason)


class WrongDatabaseSchemaOrTable(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, "Cannot fetch data via Flight Service. Try again.", reason)


class UnsupportedOutputConnection(WMLClientError, ValueError):
    def __init__(self, connection_id=None, reason=None):
        WMLClientError.__init__(self, f"Connection with ID: {connection_id} is not supported as an Output connection",
                                reason)


class UnsupportedConnection(WMLClientError, ValueError):
    def __init__(self, conn_type=None, reason=None):
        WMLClientError.__init__(self, f"Connection type: {conn_type} is not supported.",
                                reason)


class MissingFileName(WMLClientError, KeyError):
    def __init__(self, reason=None):
        WMLClientError.__init__(self, f"Connection location requires a 'file_name' to be specified.",
                                reason)


class FileUploadFailed(WMLClientError, ConnectionError):
    def __init__(self, reason=None):
        WMLClientError.__init__(self, f"Failed to upload file.",
                                reason)


class FlightServiceInfoNotAvailable(WMLClientError, ConnectionError):
    def __init__(self, reason=None):
        WMLClientError.__init__(self, "To use the data connection service from outside of Watson Studio "
                                      "you must explicitly set environment variables 'FLIGHT_SERVICE_LOCATION' "
                                      "and 'FLIGHT_SERVICE_PORT' to point to the location of the Flight Server. "
                                      "Ask your cluster administrator for the external route "
                                      "to the Flight Server for data connections.",
                                reason)



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from pandas import DataFrame
    from ibm_watson_machine_learning.helpers import DataConnection

__all__ = [
    "BaseAutoPipelinesRuns"
]


class BaseAutoPipelinesRuns:
    """
    Base abstract class for Pipeline Optimizers Runs.
    """

    @abstractmethod
    def list(self) -> 'DataFrame':
        """Lists historical runs/fits with status."""
        pass

    @abstractmethod
    def get_params(self, run_id: str = None) -> dict:
        """Get executed optimizers configs parameters based on the run_id."""
        pass

    @abstractmethod
    def get_run_details(self, run_id: str = None) -> dict:
        """Get run details. If run_id is not supplied, last run will be taken."""
        pass

    @abstractmethod
    def get_optimizer(self, run_id: str):
        """Creates instance of AutoPipelinesRuns with all computed pipelines computed by AutoAi on WML."""
        pass

    @abstractmethod
    def get_data_connections(self, run_id: str) -> List['DataConnection']:
        """Create DataConnection objects for further user usage"""
        pass



"""
IBM Confidential
OCO Source Materials
5737-H76, 5725-W78, 5900-A1R
(c) Copyright IBM Corp. 2020 All Rights Reserved.
The source code for this program is not published or otherwise divested of its trade secrets,
irrespective of what has been deposited with the U.S. Copyright Office.
"""
import json
import time


class MetricsEntry():
    """
    Stores metrics for one round. The each round has values that can be added at the respective
    hook. The hooks should only fill their values and can use any keys they like (for now).
    """

    def __init__(self):
        """
        Just initializes a dictionary for each hook and creates a variable to keep track of which
        round this entry corresponds to.
        :param: None
        """
        self.round_no = None
        self.pre_train = {}
        self.post_train = {}
        self.pre_update = {}
        self.post_update = {}

    def to_dict(self):
        """
        Converts the entry to a dictionary, simply mapping the variable names to keys.
        :return: dictionary with key, value pairs for all instance variables
        :rtype: `dict`
        """
        return {
            "round_no": self.round_no,
            "pre_train": self.pre_train,
            "post_train": self.post_train,
            "pre_update": self.pre_update,
            "post_update": self.post_update
        }


class MetricsRecorder():

    def __init__(self, output_file, output_type, compute_pre_train_eval, compute_post_train_eval):
        """
        Converts the entry to a dictionary, simply mapping the variable names to keys.
        :param output_file: path to write metrics file (without extension)
        :type output_file: `str`
        :param compute_pre_train_eval: whether to run an evaluation before local training as part of
        the train command (i.e. after the last round's sync)
        :type compute_pre_train_eval: `bool`
        :param compute_post_train_eval: whether to run an evaluation after local training as part of
        the train command (i.e. before syncing)
        :type compute_post_train_eval: `bool`
        """
        self._data = []
        self.output_file = output_file
        self.output_type = output_type
        self.compute_pre_train_eval = True
        self.compute_post_train_eval = True

    def add_entry(self):
        """
        Create a new entry, i.e. for the upcoming round.

        :param: None
        :return: None
        """
        self._data += [MetricsEntry()]

    def set_round_no(self, round_no):
        """
        Update the most recent entry with the round number.

        :param round_no: the current round number
        :type round_no: `int`
        :return: None
        """
        entry = self._data[-1]
        entry.round_no = round_no

    def pre_train_hook(self, eval_results, additional_metrics):
        """
        Fill in the pre_train dictionary of the latest entry

        :param eval_results: result from latest evaluation if compute_pre_train_eval == True \
        (else None); TODO standardize keys inside the eval_results dictionary
        :type eval_results: `dict`
        :return: None
        """
        entry = self._data[-1]
        entry.pre_train['ts'] = time.time()
        if isinstance(eval_results, dict):
            for k,v in eval_results.items():
                entry.pre_train[f'eval:{k}'] = v

        for k,v in additional_metrics:
            entry.pre_train[f'extra:{k}'] = v

    def post_train_hook(self, train_results, eval_results, additional_metrics):
        """
        Fill in the post_train dictionary of the latest entry

        :param train_results: result from the training that just took place
        TODO standardize keys inside the train_results dictionary
        :type train_results: `dict`
        :param eval_results: result from latest evaluation if compute_post_train_eval == True \
        (else None); TODO standardize keys inside the eval_results dictionary
        :type eval_results: `dict`
        :return: None
        """
        entry = self._data[-1]
        entry.post_train['ts'] = time.time()
        if isinstance(train_results, dict):
            for k,v in train_results.items():
                entry.post_train[f'train:{k}'] = v
        if isinstance(eval_results, dict):
            for k,v in eval_results.items():
                entry.post_train[f'eval:{k}'] = v

        for k,v in additional_metrics.items():
            entry.post_train[f'extra:{k}'] = v

    def pre_update_hook(self):
        """
        Fill in the pre_update dictionary of the latest entry; for now we just take timestamps

        :param: None
        :return: None
        """
        entry = self._data[-1]
        entry.pre_update['ts'] = time.time()

    def post_update_hook(self):
        """
        Fill in the post_update dictionary of the latest entry; for now we just take timestamps

        :param: None
        :return: None
        """
        entry = self._data[-1]
        entry.post_update['ts'] = time.time()

    def get_output_file(self):
        """
        Return the formatted output file, including extension

        :param: None
        :return: the full filepath to the output file
        :rtype: `str`
        """
        return '{}.{}'.format(self.output_file, self.output_type)

    def get_output_type(self):
        """
        Return the extension (and hence type) of the output, just "json" for now

        :param: None
        :return: the extension/type of how the metrics will be written to file
        :rtype: `str`
        """
        return self.output_type

    def to_json(self):
        """
        Generate json for all of the metrics data currently stored

        :param: None
        :return: json-formatted metrics data
        :rtype: `str`
        """
        entry_list = []
        for entry in self._data:
            entry_list += [entry.to_dict()]
        return json.dumps(entry_list)



# (C) Copyright IBM Corp. 2020.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Package skeleton

.. moduleauthor:: Wojciech Sobala <wojciech.sobala@pl.ibm.com>
"""

from os.path import join as path_join
import pkg_resources
import sys

try:
    wml_location = pkg_resources.get_distribution("ibm-watson-machine-learning").location
    sys.path.insert(1, path_join(wml_location, 'ibm_watson_machine_learning', 'libs'))
    sys.path.insert(2, path_join(wml_location, 'ibm_watson_machine_learning', 'tools'))
except pkg_resources.DistributionNotFound:
    pass
from ibm_watson_machine_learning.utils import version
from ibm_watson_machine_learning.client import APIClient

from .utils import is_python_2
if is_python_2():
    import logging
    logger = logging.getLogger(__name__)
    logger.warning("Python 2 is not officially supported.")
